use serde_derive::Deserialize;

#[derive(Debug, Clone, Deserialize)]
pub struct ServerConfig {
    pub host: String,
    pub port: u16,
    pub nginx_sendfile: bool,
}

#[derive(Debug, Clone, Deserialize)]
pub struct DatabaseConfig {
    pub sql_backend: String,
    pub url: String,
}

#[derive(Debug, Clone, Deserialize)]
pub struct LoggingConfig {
    pub level: String,
    pub metrics: bool,
}

#[derive(Debug, Clone, Deserialize)]
pub struct OperationsConfig {
    pub filename_size: u8,
    pub adminkey_size: u8,
    pub apikey_size: u8,
    pub engine_mode: i64,
    pub cleaner_interval: i32,
    pub file_expiry_min: u32,
    pub file_expiry_max: u32,
    pub max_filesize: u64,
    pub banned_mimetype: Vec<String>,
    pub unsafe_mimetype: Vec<String>,
}

#[derive(Debug, Clone, Deserialize, Default)]
pub struct AppConfig {
    pub server: ServerConfig,
    pub database: DatabaseConfig,
    pub logging: LoggingConfig,
    pub operations: OperationsConfig,
    pub uploading: UploadConfig,
}

#[derive(Debug, Clone, Deserialize)]
pub struct UploadConfig {
    pub allow_unverified_keys: bool,
    pub max_key_generations: i32,
    pub fallback_key: String,
}

impl Default for ServerConfig {
    fn default() -> Self {
        ServerConfig {
            host: "0.0.0.0".to_string(),
            port: 8282,
            nginx_sendfile: false,
        }
    }
}

impl Default for DatabaseConfig {
    fn default() -> Self {
        DatabaseConfig {
            sql_backend: "sqlite".to_string(),
            url: "sqlite://data/ephemeral.db".to_string(),
        }
    }
}

impl Default for LoggingConfig {
    fn default() -> Self {
        LoggingConfig {
            level: "ephemeral=info,sqlx=error,hyper=error,salvo_extra=error,salvo=error,mio=error"
                .to_string(),
            metrics: true,
        }
    }
}

impl Default for OperationsConfig {
    fn default() -> Self {
        OperationsConfig {
            filename_size: 6,
            adminkey_size: 12,
            apikey_size: 24,
            engine_mode: 2,
            cleaner_interval: 1800,
            file_expiry_min: 604800,
            file_expiry_max: 31536000,
            max_filesize: 1073741824, // 10 MB
            banned_mimetype: vec![
                "application/x-msdownload".to_string(),
                "application/x-msi".to_string(),
                "application/x-silverlight".to_string(),
                "application/x-ms-dos-executable".to_string(),
                "application/vnd.microsoft.portable-executable".to_string(),
            ],
            unsafe_mimetype: vec![
                "text/html".to_string(),
                "text/javascript".to_string(),
                "application/javascript".to_string(),
                "text/css".to_string(),
                "application/x-httpd-php".to_string(),
                "application/xhtml+xml".to_string(),
            ],
        }
    }
}

impl Default for UploadConfig {
    fn default() -> Self {
        UploadConfig {
            allow_unverified_keys: false,
            max_key_generations: 25,
            fallback_key: "set-me".to_string(),
        }
    }
}
