use sqlx::{sqlite::SqliteQueryResult, Pool, Sqlite};
use std::{collections::HashMap, time::SystemTime};
use tracing::{debug, error};

use crate::keys::Key;

/// This struct represents a single file, with a few statistics about said file instance.
pub struct FileMetric {
    pub filename: String,
    pub filesize: i64,
    pub mimetype: String,
    pub views: i64,
    pub expiry: i64,
    pub is_deleted: bool,
}

// This struct represents a single mimetype, with a few statistics about said mimetype instance
pub struct MimetypeMetric {
    pub mimetype: String,
    pub host: String,
    pub alive_files: i64,
    pub dead_files: i64,
    pub filesize_alive: i64,
    pub filesize_dead: i64,
    pub views_alive: i64,
    pub views_dead: i64,
}

// Database selection enum
#[derive(Debug, Clone, Copy, PartialEq)]
pub enum DatabaseType {
    Sqlite,
    Postgres,
    Mysql,
    Oraclecloud, // https://github.com/kubo/rust-oracle because why not?
}

// Allow you to convert a string to a database type
impl DatabaseType {
    pub fn from_str(s: &str) -> Option<Self> {
        match s {
            "sqlite" => Some(Self::Sqlite),
            "postgres" => Some(Self::Postgres),
            "mysql" => Some(Self::Mysql),
            "oraclecloud" => Some(Self::Oraclecloud),
            _ => None,
        }
    }
}

// TODO: DB Wide postgres/other database support.
// Since we can't have runtime support for multiple, we need to have compile flags until they fix the macros.
// to do that, we need to make this file just a stub, that points to the 'correct' query depending on the database that's being compiled.
// custom macros?, if statements? who knows.

/// Adding a file to the database
/// TODO: convert file from just a collection of vars to an actual struct
pub async fn add_file(
    sqlconn: &Pool<Sqlite>,
    file: &str,
    mimetype: &str,
    expiry: i32,
    // expiry_override: Option<i32>, // Optional - for future use where you can specify expiry at upload.
    owner: &str,
    last_access: i32, // Just the currenttime.
    filesize: i32,    // I dread calculating this.
    ip: &str,         // set to the end-user IP of the upload request.
    domain: &str,     // set to the HOST header of the upload request.
) -> Result<SqliteQueryResult, sqlx::Error> {
    let result = sqlx::query!(
        "INSERT INTO files (
                file,
                mimetype,
                expiry, 
                owner, 
                uploaded,
                last_access,
                filesize, 
                ip, 
                domain)
                VALUES ( ?,?,?,?,?,?,?,?,? )",
        file,
        mimetype,
        expiry,
        owner,
        last_access,
        last_access, // This one is saved as "uploaded" time.
        filesize,
        ip,
        domain
    )
    .execute(sqlconn)
    .await?;
    debug!(
        "[{}] Added {} ({}) to the database",
        &domain, &file, &mimetype
    );
    Ok(result)
    // Will need to add another else if for expiry_override if added later.
}

/// This enum represents the possible outcomes of checking if the file exists int he db.
pub enum FileCheckResult {
    Found,
    NotFound,
}

/// This function checks if the filename actually exists in the DB, returning true if it does, false otherwise.
/// This *can* return true if there are multiple files with the same name, as we're not really interested in checking for errors here.
pub async fn check_filename<'a, S>(
    sqlconn: &Pool<Sqlite>,
    filename: S,
) -> Result<FileCheckResult, sqlx::Error>
where
    S: AsRef<str> + sqlx::Encode<'a, sqlx::Sqlite>,
{
    let fname = filename.as_ref();

    let result = sqlx::query!(
        "SELECT COUNT(*) as count 
            FROM files 
            WHERE file = ? and is_deleted = 0",
        fname
    )
    .fetch_one(sqlconn)
    .await?;

    // Handle multiple files and such
    match result.count {
        0 => Ok(FileCheckResult::NotFound),
        1 => Ok(FileCheckResult::Found),
        _ => {
            // We found more than one file, so log it and say we found one.
            // I tried passing this up as MultipleFilesFound in the enum, but that made it hard to just "log and run"
            error!(
                "The check_file query received multiple entries for one file: {}, count: {}",
                &fname, result.count
            );
            Ok(FileCheckResult::Found)
        }
    }
}

/// This function checks the owner of a file.
/// Give it a filename, it'll return the owner of the file.
pub async fn check_owner(sqlconn: &Pool<Sqlite>, filename: &str) -> Result<String, sqlx::Error> {
    let row = sqlx::query!(
        "SELECT owner FROM files WHERE file = ? AND is_deleted = 0",
        filename
    )
    .fetch_one(sqlconn)
    .await?;

    let owner = row.owner;
    Ok(owner)
}

/// This returns the mimetype of the file from the database.
/// This is used for serving files, so that we can set the mimetype of unsafe files.
pub async fn get_mimetype(sqlconn: &Pool<Sqlite>, file: &str) -> Result<String, sqlx::Error> {
    let row = sqlx::query!(
        "SELECT mimetype FROM files WHERE file = ? AND is_deleted = 0",
        file
    )
    .fetch_one(sqlconn)
    .await?;

    let mimetype = row.mimetype;
    debug!("Mimetype: {}", &mimetype);
    Ok(mimetype)
}

/// This marks a file as deleted in the database.
/// This doesn't delete the file on disk, however.
/// Returns the number of the number of rows affected (files deleted).
pub async fn delete_file(sqlconn: &Pool<Sqlite>, filename: &String) -> Result<u64, sqlx::Error> {
    debug!("delete_file(adminkey: {})", filename);
    let result = sqlx::query!("UPDATE files SET is_deleted = 1 WHERE file = ?", filename)
        .execute(sqlconn)
        .await?;

    Ok(result.rows_affected())
}

/// Updates a file's viewcount and accesstime.
/// Returns a Result(u64), with the number of rows affected.
pub async fn update_fileview<S>(
    sqlconn: &Pool<Sqlite>,
    filename: S,
    last_access: i32,
) -> Result<u64, sqlx::Error>
where
    S: AsRef<str>,
{
    let filename = filename.as_ref();
    let result = sqlx::query!(
        "UPDATE files SET last_access = ?, views = views + 1 WHERE file = ?",
        last_access,
        filename
    )
    .execute(sqlconn)
    .await?;
    Ok(result.rows_affected())
}

/// Returns the unix timestamp of the last access - 0 if unviewed.
/// This will break in 2038
pub async fn get_accesss_time(sqlconn: &Pool<Sqlite>, filename: &str) -> Result<i64, sqlx::Error> {
    let result = sqlx::query!("SELECT last_access FROM files WHERE file = ?", filename)
        .fetch_one(sqlconn)
        .await?;

    Ok(result.last_access)
}

/// Generating a list of files that should be deleted
pub async fn get_old_files(sqlconn: &Pool<Sqlite>) -> Result<Vec<String>, sqlx::Error> {
    // Get the current time
    let time = SystemTime::now()
        .duration_since(SystemTime::UNIX_EPOCH)
        .unwrap()
        .as_secs() as i32;

    // Query for files where the expiry has passed (or the expiry_override has passed, if set.)
    // Also only returns files that haven't been deleted yet.
    let result = sqlx::query!(
        "SELECT file FROM files WHERE (expiry_override <= ? AND expiry_override NOT NULL) OR (expiry <= ?) AND is_deleted = 0",
        time,
        time
    )
    .fetch_all(sqlconn)
    .await?;

    // Parse the result into the vec.
    // This 'borrows' the result directly from the result, so their lifetime is tried to the lifetime of the database connection.
    // This is fine since it's actual lifetime is short compared.
    let files: Vec<String> = result.iter().map(|r| r.file.to_owned()).collect();

    // Print debug and return
    debug!("get_old_files(files: {:?})", files);
    Ok(files)
}

/// This adds a new QR code scan to the database.
/// The version string is the 'iteration' of the qr code itself, as I'd like to print multiple qr codes with multiple features.
/// TODO: we should probably move the time out of this function, so we can change the time if backfilling.
pub async fn add_qrscan(
    sqlconn: &Pool<Sqlite>,
    ip: String,
    useragent: String,
    version: String,
) -> Result<SqliteQueryResult, sqlx::Error> {
    let time = SystemTime::now()
        .duration_since(SystemTime::UNIX_EPOCH)
        .unwrap()
        .as_secs() as i32;
    sqlx::query!(
        "INSERT INTO qrscan (
                time,
                IP,
                useragent, 
                version)
                VALUES ( ?,?,?,? )",
        time,
        ip,
        useragent,
        version
    )
    .execute(sqlconn)
    .await
}

// /// Updating the expiry_override of a file.
// /// There are a few limitations and considerations for expiry_override:
// ///
// /// 1. The expiry_override must be set lower than the expiry of the file.
// /// 2. EXCEPT when the file has been uploaded by an API Key that has permission to change expiry_override beyond the expiry
// /// 3. The expiry_override cannot be set longer than file_expiry_max, unless the file has been uploaded by an API Key that allows 'permanent' uploads.
// ///
// /// If /any/ of those limits are broken, this function will not update and return an error, with the reason why.
// fn update_expiry_override() {
//     todo!()
// }

/// This function gets a Vec of all the files uploaded by a specific API Key.
/// This allows for the select few who hold an API key to view the information about the files they upload.
pub async fn get_files_by_key(
    sqlconn: &Pool<Sqlite>,
    owner: &Key,
) -> Result<Vec<FileMetric>, sqlx::Error> {
    let result = sqlx::query!(
        "SELECT file, filesize, mimetype, views, expiry, expiry_override, ip, is_deleted
            FROM files
            WHERE owner = ?",
        owner.uuid
    )
    .fetch_all(sqlconn)
    .await?;

    let mut files: Vec<FileMetric> = Vec::new();

    for row in result {
        // Convert the is_deleted from a int to a bool
        let is_deleted = row.is_deleted.map(|v| v != 0).unwrap_or(false);

        // For each file in the result, make a new FileMetric object and add it to the list
        let file = FileMetric {
            filename: row.file,
            mimetype: row.mimetype,
            filesize: row.filesize,
            views: row.views.unwrap(),
            expiry: row.expiry,
            is_deleted,
        };
        files.push(file);
    }
    Ok(files)
}

/// Returns how many files exist on the server.
pub async fn file_count(sqlconn: &Pool<Sqlite>) -> Result<i32, sqlx::Error> {
    let count = sqlx::query!(
        r#"
            SELECT COUNT(*) as count FROM files
        "#
    )
    .fetch_one(sqlconn)
    .await?;

    Ok(count.count)
}

// Globally important stats.

/// This counts the total files that have been uploaded per ip address.
/// This returns a hashmap<String, i32> containing the IP, and the number of files uploaded.
/// This isn't perfect, as some devices that use IPv6 can sometimes use that, but then also switch to IPv4.
pub async fn get_total_uploads_ip(sqlconn: &Pool<Sqlite>) -> Option<HashMap<String, i32>> {
    // SELECT ip, COUNT(file) as count FROM files GROUP BY ip ORDER BY COUNT;
    let result = sqlx::query!(
        "SELECT ip, COUNT(file) as filecount
            FROM files 
            GROUP BY ip 
            ORDER BY filecount",
    )
    .fetch_all(sqlconn)
    .await;

    let mut ipcount: HashMap<String, i32> = HashMap::new();

    if result.is_err() {
        // If Error, return none and log.
        error!("Problem getting total files by ip: {:?}", result);
        None
    } else {
        for row in result.unwrap() {
            ipcount.insert(row.IP, row.filecount);
        }
        Some(ipcount)
    }
}

/// This queries the db for the total filesize (in bytes) of all alive files.
/// This can VERY Quickly become massive, so we use a u128.
///
/// 340,282,366,920,938,463,463,374,607,431,768,211,455 bytes
/// 3.095×10^26 TiB, or 309500000000000000000000000 TiB
/// or 3700000000000000000000 * more data than exists on the internet (as of 2014).
pub async fn total_alive_filesize(sqlconn: &Pool<Sqlite>) -> Option<u128> {
    let mut total_filesize: u128 = 0;
    let result = sqlx::query!(
        "SELECT filesize
            FROM files 
            WHERE is_deleted == 0"
    )
    .fetch_all(sqlconn)
    .await;
    if result.is_err() {
        // If Error, return none and log.
        error!("Problem getting total files by ip: {:?}", result);
        None
    } else {
        for row in result.unwrap() {
            total_filesize += row.filesize as u128;
        }
        Some(total_filesize)
    }
}

/// This queries the db for the total filesize (in bytes) of all dead files.
/// This can VERY Quickly become massive, so we use a u128.
///
/// 340,282,366,920,938,463,463,374,607,431,768,211,455 bytes
///
/// 3.095×10^26 TiB, or 309500000000000000000000000 TiB
///
/// or 3700000000000000000000 times more data than exists on the internet (as of 2014).
pub async fn total_dead_filesize(sqlconn: &Pool<Sqlite>) -> Option<u128> {
    let mut total_dead_filesize: u128 = 0;
    let result = sqlx::query!(
        "SELECT filesize
            FROM files 
            WHERE is_deleted == 1"
    )
    .fetch_all(sqlconn)
    .await;
    if result.is_err() {
        // If Error, return none and log.
        error!("Problem getting total files by ip: {:?}", result);
        None
    } else {
        for row in result.unwrap() {
            total_dead_filesize += row.filesize as u128;
        }
        Some(total_dead_filesize)
    }
}

/// This queries the db for the filesize for /each/ alive file. - We won't need to do this for dead files
/// since they were alive at some point, and when they are removed from the alive list, we don't really care.
/// This returns a Hashmap of the filename, filetype, and filesize inside a Vector. (How messy)
/// We want to group them by file, or filetype, or total in grafana, so it needs to be run on each file.
pub async fn get_file_metrics(sqlconn: &Pool<Sqlite>) -> Option<Vec<FileMetric>> {
    let result = sqlx::query!(
        "SELECT file, filesize, mimetype, views, expiry
        FROM files 
        WHERE is_deleted == 0",
    )
    .fetch_all(sqlconn)
    .await;

    let mut filevec: Vec<FileMetric> = Vec::new();

    if result.is_err() {
        // If Error, return none and log.
        error!("Problem getting filesizes: {:?}", result);
        None
    } else {
        for row in result.unwrap() {
            // For each row (file), add it to the struct
            let file = FileMetric {
                filename: row.file,
                mimetype: row.mimetype.clone(),
                filesize: row.filesize,
                views: row.views.unwrap(),
                expiry: row.expiry,
                is_deleted: false,
            };
            // Then add the struct to the Vec
            filevec.push(file);
        }
        // Return vec or return nothing.
        Some(filevec)
    }
}

/// This function queries the db for metrics related to mimetypes.
/// Returns a Vec of if the MimetypeMetric Struct.
pub async fn get_mimetype_metrics(
    sqlconn: &Pool<Sqlite>,
) -> Result<Vec<MimetypeMetric>, sqlx::Error> {
    // Get stats about the files
    // This is insane, ChatGPT wrote this entire query. I did no changes at all and it just worked.
    // I cannot wait until it takes our jobs.
    let result = sqlx::query!(
        "SELECT mimetype, domain, 
            SUM(CASE WHEN is_deleted = 0 THEN views ELSE 0 END) AS alive_views, 
            SUM(CASE WHEN is_deleted = 1 THEN views ELSE 0 END) AS dead_views, 
            SUM(CASE WHEN is_deleted = 0 THEN 1 ELSE 0 END) AS alive_files, 
            SUM(CASE WHEN is_deleted = 1 THEN 1 ELSE 0 END) AS dead_files, 
            SUM(CASE WHEN is_deleted = 0 THEN filesize ELSE 0 END) AS alive_filesize,
            SUM(CASE WHEN is_deleted = 1 THEN filesize ELSE 0 END) AS dead_filesize
        FROM files
        GROUP BY mimetype, domain;",
    )
    .fetch_all(sqlconn)
    .await?;

    let mimevec = result
        .iter()
        .map(|row| MimetypeMetric {
            mimetype: row.mimetype.clone(),
            host: row.domain.clone(),
            alive_files: row.alive_files as i64,
            dead_files: row.dead_files as i64,
            filesize_alive: row.alive_filesize as i64,
            filesize_dead: row.dead_filesize as i64,
            views_alive: row.alive_views.unwrap_or(0) as i64,
            views_dead: row.dead_views.unwrap_or(0) as i64,
        })
        .collect::<Vec<MimetypeMetric>>();

    Ok(mimevec)
}
