use std::collections::HashMap;

use chrono::{DateTime, Utc};
use chrono_humanize::{Accuracy, HumanTime, Tense};
use salvo::Writer;
use salvo::{
    handler,
    hyper::header::HOST,
    oapi::{endpoint, extract::QueryParam},
    prelude::StatusCode,
    writing::Text,
    Depot, Request, Response,
};
use tracing::{error, info, warn};

use crate::db;
use crate::handlers::guess_ip;
use crate::{
    handlers::{render_template, Generic, TemplateEnum},
    keys::{
        db_keys, {Key, KeyLevel},
    },
    SQLITE,
};

use super::Admin;

/// This generates the admin page that allows admins to view all keys, edit them, and investigate files uploaded by them.
#[handler]
pub async fn admin(req: &mut Request, res: &mut Response, depot: &mut Depot) {
    let headers = req.headers().clone();
    let remote_addr = &req.remote_addr().clone();
    let sqlconn = SQLITE.get().unwrap();

    let client_key = depot
        .get::<Key>("client_key")
        .expect("Failed to grab API Key from depot")
        .clone();

    // We should make sure the user is an admin, since the auth middleware only checks if the key isn't revoked.
    if client_key.key_level == KeyLevel::Master {
        let key_list = db_keys::get_all_keys(sqlconn).await.unwrap();

        let key_count_result = db_keys::count_keys(sqlconn).await;
        let key_count = match key_count_result {
            Ok(count) => count,
            Err(e) => {
                error!("Failed to count keys: {}", e);
                res.status_code(StatusCode::INTERNAL_SERVER_ERROR);
                res.render(Text::Json(r#"{"error": "CountKeysFailed"}"#));
                return;
            }
        };

        let file_count_result = db::file_count(sqlconn).await;
        let file_count = match file_count_result {
            Ok(count) => count,
            Err(e) => {
                error!("Failed to count files: {}", e);
                res.status_code(StatusCode::INTERNAL_SERVER_ERROR);
                res.render(Text::Json(r#"{"error": "CountFilesFailed"}"#));
                return;
            }
        };

        let tree = construct_tree(key_list);

        let html = generate_html_tree(&tree);

        let template_filename = "admin.html";
        let template = TemplateEnum::Admin(Admin {
            domain: String::from(headers[HOST].to_str().unwrap()),
            apikey_count: key_count.to_string(),
            file_count: file_count.to_string(),
            key_tree: html,
        });
        let status_code = StatusCode::OK;

        render_template(res, &headers, template_filename, template, status_code).await;
        return;
    } else {
        info!(
            "[{:?} - {:?}] User attempted to view '/admin' without a valid master key. Blocking.",
            headers[HOST].to_str().unwrap(),
            guess_ip(&headers, remote_addr)
        );
        let template_filename = "error.html";
        let template = TemplateEnum::Generic(Generic {
            domain: String::from(headers[HOST].to_str().unwrap()),
            message1: String::from("Error 403"),
            message2: String::from("You don't seem very 'adminy' to me."),
            ..Default::default()
        });
        let status_code = StatusCode::FORBIDDEN;
        render_template(res, &headers, template_filename, template, status_code).await;
        return;
    }
}

/// Struct for handling the hashmap.
pub struct TreeNode {
    pub key: Key,
    pub children: Vec<TreeNode>,
}

/// Constructs the tree used for the admin page.
/// This might take quite a bit of processing and ram. Who really knows though?
/// Imagine if sqlx didn't break when I tried adding references to other keys for their children.
/// That would be truly amazing, and it would be nowhere near as complicated as this.
fn construct_tree(keys: Vec<Key>) -> Vec<TreeNode> {
    // Create a HashMap that stores the keys.
    let mut key_map = HashMap::new();

    // Populate the HashMap with key UUID mapping to the actual Key struct
    for key in keys.into_iter() {
        key_map.insert(key.uuid.clone(), key);
    }

    // Initialise a (empty) List of nodes with no parents.
    let mut root_nodes: Vec<TreeNode> = Vec::new();

    // Recursive functions? You can do that??
    fn build_tree_node(
        key: Key,
        key_map: &HashMap<String, Key>,
        children_map: &HashMap<String, Vec<String>>,
    ) -> TreeNode {
        // Get the list of children UUIDs.
        let binding = Vec::new();
        let children_uuids = children_map.get(&key.uuid).unwrap_or(&binding);

        // Build the children recursively.
        let children: Vec<TreeNode> = children_uuids
            .iter()
            .filter_map(|uuid| key_map.get(uuid))
            .map(|child_key| build_tree_node(child_key.clone(), key_map, children_map))
            .collect();

        // Return a TreeNode with the key itself and its children.
        TreeNode { key, children }
    }

    // Create a HashMap for child UUIDs
    let mut children_map = HashMap::new();
    for key in key_map.values() {
        let child_uuids: Vec<String> = key.child_uuids.clone(); // Assume `Key` struct now has `child_uuids: Vec<String>`
        children_map.insert(key.uuid.clone(), child_uuids);
    }

    // Iterate on all keys again
    for key in key_map.values() {
        // Find the children of keys with no parents (master keys)
        if key.parent_key.is_none() {
            root_nodes.push(build_tree_node(key.clone(), &key_map, &children_map));
        }
    }

    // root_nodes now contains the full forest of trees
    root_nodes
}

// Top-level function to generate the entire HTML tree
fn generate_html_tree(tree: &[TreeNode]) -> String {
    let mut html = String::new();
    html.push_str("<ul id=\"hide_bullets\">");
    for root in tree {
        html.push_str(&generate_html(root));
    }
    html.push_str("</ul>");
    html
}

// Helper function to recursively generate HTML for the tree
fn generate_html(node: &TreeNode) -> String {
    // Humanize the time.
    let key_creation = {
        let datetime: DateTime<Utc> =
            DateTime::from_timestamp(node.key.creation_time.into(), 0).unwrap();
        let human_time = HumanTime::from(datetime);
        human_time.to_text_en(Accuracy::Rough, Tense::Past)
    };

    let details = format!(
        "🔑 {} 🔒 {}{} 🕛 {}{}",
        node.key.uuid,
        node.key.key_level,
        if let Some(ref comment) = node.key.comment {
            format!(
                " <span class=\"editable\" data-key=\"{}\">💬 {}</span>",
                node.key.uuid, comment
            )
        } else {
            format!(
                " <span class=\"editable\" data-key=\"{}\">💬 Add Comment</span>",
                node.key.uuid
            )
        },
        key_creation,
        if let Some(ref revoked_reason) = node.key.revoked {
            format!(" 🚫 {}", revoked_reason)
        } else {
            "".to_string()
        }
    );

    // Grab the number of children a key has.
    let child_count = node.children.len();

    // Add the start of the key.
    let mut html = "<li class=\"key-node\">".to_string();

    // If there's children keys, show them.
    html.push_str(&format!(
        "<span class=\"toggle\" collapsed=\"{}\">{}</span> ",
        if child_count > 0 { "false" } else { "none" },
        if child_count > 0 { "▼" } else { "" }
    ));

    // Show the details of keys.
    html.push_str(&details.to_string());

    // Add delete button
    html.push_str(&format!(
        " <span class=\"delete\" data-key=\"{}\">❌</span>",
        node.key.uuid
    ));

    // If there's children, make a new list and call this again on them.
    if child_count > 0 {
        html.push_str("<ul class=\"children\" style=\"display: block;\">");
        for child in &node.children {
            html.push_str(&generate_html(child));
        }
        html.push_str("</ul>");
    }

    // Close the list item.
    html.push_str("</li>");

    html
}

/// Handler to process key revocation via the API.
#[endpoint(
    tags("admin"),
    parameters(
        ("apikey", description = "The API key you're about to ban."),
        ("revoke_message", description = "The 'ban message' to show the user when they try to interact with the website."),
        ("recursive", description = "Do we also delete all children keys?"),
    )
)]
pub async fn admin_revoke_key(
    req: &mut Request,
    res: &mut Response,
    depot: &mut Depot,
    revoke_message: QueryParam<String, true>,
    recursive: QueryParam<bool, false>,
) {
    // Base Stuffs
    let sqlconn = SQLITE.get().unwrap();
    let client_key = depot
        .get::<Key>("client_key")
        .expect("Failed to grab API Key from depot")
        .clone();

    let revoke_key: String = req.param("apikey").expect("Invalid API Key.");

    // Make sure the user is a master key.
    if client_key.key_level == KeyLevel::Master {
        // Parse the key and message from the request.
        let revoke_message = revoke_message.as_str();

        // was the delete recursive, if so we want to grab all the children, ban them too.
        match recursive.into_inner().unwrap_or(false) {
            true => {
                // Build a vector of all children keys. Recursively.
                // let mut keys_to_revoke = [revoke_key.clone()];
                todo!()
            }
            false => {
                // Just delete the provided key.
                match db_keys::revoke_key(sqlconn, &revoke_key, revoke_message).await {
                    Ok(keys_revoked) => {
                        // 0 keys revoked must mean it wasn't a key.
                        if keys_revoked == 0 {
                            res.status_code(StatusCode::OK);
                            res.render(Text::Json("{\"error\": \"NoKeyFound\"}"));
                            return;
                        } else {
                            res.status_code(StatusCode::OK);
                            res.render(Text::Json("{\"success\": \"KeyRevoked\"}"));
                            return;
                        }
                    }
                    Err(e) => {
                        error!("Error while revoking key: {}", e);
                        res.status_code(StatusCode::INTERNAL_SERVER_ERROR);
                        res.render(Text::Json(r#"{"error": "RevokeKeyFailed"}"#));
                        return;
                    }
                };
            }
        }
    } else {
        res.status_code(StatusCode::FORBIDDEN);
        res.render(Text::Json(r#"{"error": "Unauthorized"}"#));
    }
}

/// This handler process editing key comments via the API.
#[endpoint(
    tags("admin"),
    parameters(
        ("apikey", description = "The API key you're about to edit."),
        ("comment", description = "The new comment to set for the provided API Key."),
    )
)]
pub async fn admin_edit_comment(
    req: &mut Request,
    res: &mut Response,
    depot: &mut Depot,
    comment: QueryParam<String, true>,
) {
    // Base Stuffs
    let sqlconn = SQLITE.get().unwrap();
    let client_key = depot
        .get::<Key>("client_key")
        .expect("Failed to grab API Key from depot")
        .clone();

    let edit_key: String = req
        .param("apikey")
        .expect("Failed to set the filename variable");

    // Make sure the user is a master key.
    if client_key.key_level == KeyLevel::Master {
        let comment = comment.into_inner();

        // Update the comment for the provided key.
        match db_keys::edit_comment(sqlconn, &edit_key, &comment).await {
            Ok(success) => {
                if success {
                    res.status_code(StatusCode::OK);
                    info!(
                        "[ADMIN] Updated comment of key ({}) to: {}",
                        &edit_key, &comment
                    );
                    res.render(Text::Json("{\"success\": \"CommentUpdated\"}"));
                } else {
                    res.status_code(StatusCode::OK);
                    warn!("Cosmic Ray Detected, impossible we got to an impossible path!");
                    res.render(Text::Json("{\"error\": \"UpdateFailed\"}"));
                }
            }
            Err(e) => {
                res.status_code(StatusCode::INTERNAL_SERVER_ERROR);
                error!("Error while editing comment: {}", e);
                res.render(Text::Json(r#"{"error": "EditCommentFailed"}"#));
            }
        }
    } else {
        res.status_code(StatusCode::FORBIDDEN);
        res.render(Text::Json(r#"{"error": "Unauthorized"}"#));
    }
}
