use salvo::{
    http::header::{self, HOST},
    prelude::*,
};
use tracing::{debug, error, info};

use crate::{
    handlers::{guess_ip, is_browser, render_template, Generic, TemplateEnum},
    keys::{self, db_keys},
    SQLITE,
};

/// This handles authentication as a middleware.
/// We accept Authorization: Bearer <token> keys. While it's not directly OAuth2 compliant, it'll work.
/// If unverified_keys is true, we essentially just say "all keys are valid"
/// The Depot stores the Key (object, including the uuid), ip and host in the Depot, so we can check it later when handling the actual requests.
#[handler]
pub async fn require_auth(res: &mut Response, req: &mut Request, depot: &mut Depot) {
    // Initialise the basic stuff we need
    // Clone out the header and remote_addr so we can guess the ip later
    let headers = &req.headers().clone();
    let remote_addr = &req.remote_addr().clone();
    // Get the host header for nicely setting up the response.
    debug!("require_auth(req): {:?}", req);

    // Setup the Sqlite pool stuff for later
    let sqlconn = SQLITE.get().unwrap();
    let is_user_web = is_browser(req).await;
    // Debug print the headers found.
    for header in headers.iter() {
        debug!("header: {:?}", header);
    }
    // Debug print cookies found.
    for cookie in req.cookies().iter() {
        debug!("cookie: {:?}", cookie)
    }
    // Check for a token in the header.
    let header_key: Option<&str> = req.header(header::AUTHORIZATION);
    // Split token into apikey if present.
    let token_from_header = match header_key {
        Some(token) => token.strip_prefix("Bearer "),
        None => None,
    };
    // Check, and map the token from cookie.
    let token_from_cookie: Option<&str> = req.cookies().get("api-key").map(|cookie| cookie.value());
    // Let the token be either one of these.
    let token = token_from_cookie.or(token_from_header);
    if token.is_none() {
        info!(
            "[{:?} - {:?}] Attempted Access of authorised endpoint. Blocking.",
            headers[HOST].to_str().unwrap(),
            guess_ip(headers, remote_addr)
        );
        // Key is not Present. Failing.
        if is_user_web {
            // We need to check if they're a web user, since this middleware protects both API and frontend pages.
            let template_filename = "error.html";
            let template = TemplateEnum::Generic(Generic {
                domain: String::from(headers[HOST].to_str().unwrap()),
                message1: String::from("Error 401"),
                message2: String::from("This server is configured to require authentication for this action. Did you forget your API Key?"),
                ..Default::default()
            });
            let status_code = StatusCode::FORBIDDEN;
            render_template(res, headers, template_filename, template, status_code).await;
            return;
        } else {
            // Otherwise, render the error in json for the API.
            res.status_code(StatusCode::FORBIDDEN);
            res.render(Text::Json(r#"{"error": "MissingAPIKey"}"#));
            return;
        }
    } else {
        // Retrieve the Key from the database, if it exists.
        let key_result =
            db_keys::get_key(sqlconn, token.expect("Somehow there were no tokens.")).await;
        // Do the actual check to handle possible outcomes
        match key_result {
            Ok(valid_key) => {
                // Key exists.

                // Check if the key (from key_result) has been revoked.
                if let Some(revoke_message) = valid_key.revoked {
                    // Banned, show ban message.
                    info!(
                        "[{:?} - {:?}] Attempted File upload with revoked API Key. Blocking.",
                        headers[HOST].to_str().unwrap(),
                        guess_ip(headers, remote_addr)
                    );
                    // Check if the upload came from the web
                    if is_user_web {
                        let template_filename = "error.html";
                        let template = TemplateEnum::Generic(Generic {
                            domain: String::from(headers[HOST].to_str().unwrap()),
                            message1: String::from("Key Revoked"),
                            message2: revoke_message,
                            ..Default::default()
                        });

                        let status_code = StatusCode::FORBIDDEN;

                        render_template(res, headers, template_filename, template, status_code)
                            .await;
                        return;
                    } else {
                        // Otherwise, render the error in json for the API.
                        res.status_code(StatusCode::FORBIDDEN);
                        res.render(Text::Json(format!(
                            r#"{{"error": "APIKeyRevoked", "reason": "{}"}}"#,
                            revoke_message
                        )));
                        return;
                    }
                } else {
                    // Not banned, push the key object to the depot and continue on with the upload.
                    // Clone it so we can pass it to the debug. That's literally it. The function ends here so it's not really a performance problem. (I hope)
                    depot.insert("client_key", valid_key.clone());
                    debug!(
                        "Key {} @ level: {:?} passed the vibe check. Proceeding to page",
                        &valid_key.uuid, &valid_key.key_level
                    );
                    // Return out of the function, we've done all the checks! \o/
                    return;
                }
            }
            Err(keys::db_keys::KeyError::NotFound) => {
                // Key not found.
                info!(
                    "[{:?} - {:?}] Attempted Access with an invalid API Key ({}). Blocking.",
                    headers[HOST].to_str().unwrap(),
                    guess_ip(headers, remote_addr),
                    token.expect("Somehow failed to unwrap token.")
                );
                // Check if the upload came from the web
                if is_user_web {
                    let template_filename = "error.html";
                    let template = TemplateEnum::Generic(Generic {
                        domain: String::from(headers[HOST].to_str().unwrap()),
                        message1: String::from("Error 401"),
                        message2: String::from("You're using an invalid API Key. Sorry."),
                        ..Default::default()
                    });

                    let status_code = StatusCode::UNAUTHORIZED;

                    render_template(res, headers, template_filename, template, status_code).await;
                    return;
                } else {
                    // Otherwise, render the error in json for the API.
                    res.status_code(StatusCode::UNAUTHORIZED);
                    res.render(Text::Json(r#"{"error": "InvalidAPIKey"}"#));
                    return;
                }
            }
            Err(keys::db_keys::KeyError::DatabaseError(e)) => {
                // Error when reading the database.
                error!("Error when reading the database: {:?}", e);
                // Check if the upload came from the web
                if is_user_web {
                    let template_filename = "error.html";
                    let template = TemplateEnum::Generic(Generic {
                        domain: String::from(headers[HOST].to_str().unwrap()),
                        message1: String::from("Error 500"),
                        message2: String::from("Database error when matching client key."),
                        ..Default::default()
                    });

                    let status_code = StatusCode::SERVICE_UNAVAILABLE;

                    render_template(res, headers, template_filename, template, status_code).await;
                    return;
                } else {
                    // Otherwise, render the error in json for the API.
                    res.status_code(StatusCode::SERVICE_UNAVAILABLE);
                    res.render(Text::Json(r#"{"error": "DatabaseFailure"}"#));
                    return;
                }
            }
        }
    }
}
