use salvo::{
    hyper::header::HOST, oapi::endpoint, prelude::StatusCode, writing::Text, Depot, Request,
    Response,
};
use tracing::{debug, error, info};

use crate::{db, engine, handlers::guess_ip, keys::Key, SQLITE};

/// Endpoint for deleting files.
///
/// This permanently deletes the physical file from disk, and marks the file deleted in the database.
#[endpoint(
    tags("file-management"),
    parameters(
        ("filename", description = "The name of the file you're about to delete."),
    )
)]
pub async fn delete_file(req: &mut Request, res: &mut Response, depot: &mut Depot) {
    // Set up all the things we need for the function later.
    let headers = req.headers().clone();
    let remote_addr = &req.remote_addr().clone();
    let sqlconn = SQLITE.get().unwrap();
    let filename: &String = &req.param("filename").unwrap();

    debug!("Received request to delete file: {}", filename);

    // Grab the user key from the auth handler.
    let client_key = depot
        .get::<Key>("client_key")
        .expect("Failed to grab API Key from depot")
        .clone();

    // Grab the file's owner key from the db.
    let owner = db::check_owner(sqlconn, filename).await;

    debug!(
        "Deletion attempt: user: {} file owner: {:?}",
        client_key.uuid, owner
    );
    // Compare the client_key and owner.
    let authorized = match owner {
        Ok(owner_key) => client_key.uuid == owner_key,
        Err(_) => false,
    };
    debug!("authorized: {}", authorized);

    // Proceed with deletion if authorized, otherwise respond with an error.
    if authorized {
        // Actually delete the file from disk
        db::delete_file(sqlconn, filename)
            .await
            .unwrap_or_else(|err| {
                error!(
                    "Failed to delete file from the database: {} error: {:?}",
                    filename, err
                );
                0
            });

        engine::delete_file(&filename).await.unwrap_or_else(|err| {
            error!(
                "Failed to delete file from the filesystem: {} error: {:?}",
                filename, err
            );
        });

        info!(
            "[{:?} - {:?}] User deleted file '{}'",
            headers[HOST].to_str().unwrap(),
            guess_ip(&headers, remote_addr),
            filename
        );

        res.status_code(StatusCode::OK);
        res.render(Text::Json(r#"{"deletion": "success"}"#));

    // Invalid API Key. The API Exists, but isn't right.
    } else {
        res.status_code(StatusCode::UNAUTHORIZED);
        res.render(Text::Json(r#"{"error": "InvalidAPIKey"}"#));
    }
}
