use std::{
    fs::{self, File},
    io::Read,
    path::PathBuf,
    time::{Duration, SystemTime},
};

use chrono::{DateTime, Utc};
use chrono_humanize::Humanize;
use ramhorns::{Content, Ramhorns};
use salvo::{
    conn::SocketAddr,
    hyper::{header::HOST, HeaderMap},
    prelude::StatusCode,
    writing::Text,
    Request, Response,
};
use sqlx::{sqlite::SqliteQueryResult, Error};
use tokio::{task, time};
use tracing::{debug, error, info};

use crate::{
    db::{self, FileMetric},
    engine::{self, calculate_expiry},
    CONFIG, COOKIE, SQLITE,
};

// Submodules for each request handler
pub mod admin;
pub mod auth;
pub mod delete_file;
pub mod index;
pub mod serve_file;
pub mod serve_metrics;
pub mod serve_static;
pub mod upload;
pub mod user_profile;

/// This file contains sub-modules that are the handlers for each type of web request.
/// Thanks audron for making me split my work up into multiple files.
/// like seriously this /is/ more clean
/// https://stackoverflow.com/questions/26435102/in-rust-what-is-the-purpose-of-a-mod-rs-file
/// This enum stores all valid Template Structs
pub enum TemplateEnum {
    Generic(Generic),
    Profile(Profile),
    Admin(Admin),
}

/// This struct stores the custom data shown on most pages.
#[derive(Content, Default)]
pub struct Generic {
    domain: String,
    fileurl: String,
    adminurl: String,
    message1: String,
    message2: String,
}

/// This struct stores the information shown to the user when browsing the /profile page.
#[derive(Content, Default)]
pub struct Profile {
    domain: String,         // The domain the user is visiting from.
    key_creation: String,   // A nice humanised time when the key was created.
    key_generator: String,  // The HTML for the Key Generator frontend
    key_list_html: String,  // The HTML for listing keys.
    file_list_html: String, // The HTML for listing files.
}

/// This struct stores the information shown to the user when browsing the /profile page.
#[derive(Content, Default)]
pub struct Admin {
    domain: String,       // The domain the user is visiting from.
    apikey_count: String, // How many API Keys exist.
    file_count: String,   // How many files exist.
    key_tree: String,     // A nice humanised time when the key was created.
}

// Import tests for handlers
#[cfg(test)]
mod mod_tests;

/// This function returns what the best guess of the Client's IP address is.
///
///
/// This isn't perfect, it won't magically figure out VPN's or proxies,
/// so don't use this for authentication or anything.
///
/// # Examples
///
/// ```
/// // Clone out the header and remote_addr so we can guess the ip later
/// let headers = &req.headers().clone();
/// let remote_addr = &req.remote_addr().unwrap().clone();
/// let ip = engine::guess_ip(headers, remote_addr);
/// ```
pub fn guess_ip(headers: &HeaderMap, ip: &SocketAddr) -> String {
    // Define cloudflare ip, or None
    let cloudflare_ip: Option<String> = headers
        .get("CF-Connecting-IP")
        .map(|hv| hv.to_str().unwrap().to_owned());
    debug!("guess_ip(CF-Connecting-IP: {:?})", cloudflare_ip);
    // Define x_real_ip, or None
    let x_real_ip: Option<String> = headers
        .get("X-Real-IP")
        .map(|hv| hv.to_str().unwrap().to_owned());
    debug!("guess_ip(X-Real-IP: {:?})", x_real_ip);

    if cloudflare_ip.is_none() {
        // Either there was a problem with the cloudflare header, or there is no cloudflare.
        debug!("guess_ip(No cloudflare header detected!, falling back to X-Real-IP header)");
        if x_real_ip.is_none() {
            debug!("guess_ip(No X-Real-IP detected!, falling back to remote_addr header)");

            match ip {
                SocketAddr::IPv4(addr) => addr.ip().to_string(),
                SocketAddr::IPv6(addr) => addr.ip().to_string(),
                SocketAddr::Unix(_) => "0.0.0.0".to_string(),
                _ => "0.0.0.0".to_string(),
            }
        } else {
            x_real_ip.unwrap_or("0.0.0.0".to_string())
        }
    } else {
        cloudflare_ip.unwrap_or("0.0.0.0".to_string())
    }
}

// Template rendering function
// When we want to render a page, we just pass over the TemplateStruct and anything else.

// I'm thinking of improving this a bit, where /all/ rendering is passed through this.
// All error handling and all branches of potential WEB and API passes through here.
// This function should be self-contained, so all it should receive is the following
// res, for building the response, and adding to any previous changes
// headers, for grabbing the domain and building the template path.
// template_filename, to tell the renderer what 'template' to render.
// TemplateStruct, which is a custom struct that contains all the variables that end up on the webpage.
pub async fn render_template(
    res: &mut Response,
    headers: &HeaderMap,
    template_filename: &str,
    template: TemplateEnum,
    status_code: StatusCode,
) {
    // Parse the host from the headers
    let host = headers
        .get(HOST)
        .and_then(|h| h.to_str().ok())
        .unwrap_or("localhost:8282");

    // Build the template path from the host
    let template_host_path = PathBuf::from("./templates/").join(host);

    // Load the template
    let tpls: Ramhorns = match Ramhorns::from_folder(template_host_path) {
        Ok(tpls) => tpls,
        Err(_) => {
            // If the template is not found, fall back to the default template file
            let default_template_host_path = PathBuf::from("./templates/localhost:8282/");
            Ramhorns::from_folder(default_template_host_path).expect("Couldn't find fallback template directory. Please ensure that the './template/localhost:8282' directory exists and is accessible.")
        }
    };

    // Tell the engine to render the template assigned above
    // Render the template based on the given TemplateEnum
    let rendered = match template {
        TemplateEnum::Generic(data) => tpls
            .get(template_filename)
            .expect("Couldn't find the template filename")
            .render(&data),
        TemplateEnum::Profile(data) => tpls
            .get(template_filename)
            .expect("Couldn't find the template filename")
            .render(&data),
        TemplateEnum::Admin(data) => tpls
            .get(template_filename)
            .expect("Couldn't find the template filename")
            .render(&data),
    };

    // Set the status code and render the completed page
    res.status_code(status_code);
    res.render(Text::Html(rendered));
}

/// This checks if the given mimetype is allowed or not.
/// 1 = banned, 0 = allowed
async fn is_mimetype_banned<S>(mimetype: S) -> bool
where
    S: AsRef<str>,
{
    // Load the banned files from the config
    let banned = &CONFIG.operations.banned_mimetype;

    // Convert the provided mimetype to lowercase for case-insensitive comparison
    let lowercase_mimetype = mimetype.as_ref().to_lowercase();

    // Convert the banned mimetypes to lowercase and collect into a Vec
    let lowercase_banned: Vec<String> = banned
        .iter()
        .map(|mime| {
            let mime_str = mime.clone().to_lowercase();
            debug!("Banned mimetype: {}", mime_str);
            mime_str
        })
        .collect();

    debug!("Provided mimetype: {}", lowercase_mimetype);

    lowercase_banned.contains(&lowercase_mimetype)
}

/// Convert the filesize to a human readable number
fn convert_file_size(file_size: i64) -> String {
    let units = ["B", "KB", "MB", "GB", "TB"];
    let base = 1024; // Base for conversion (1024 bytes = 1 kilobyte)

    if file_size < base {
        return format!("{} {}", file_size, units[0]);
    }

    let exponent = (file_size as f64).log(base as f64).floor() as u32;
    let converted_size = (file_size as f64 / base.pow(exponent) as f64).round();

    format!("{:.2} {}", converted_size, units[exponent as usize])
}

/// Convert unix timestamp to a human friendly timestamp
fn convert_unix_timestamp(unix_timestamp: i64) -> String {
    let datetime: DateTime<Utc> = DateTime::from_timestamp(unix_timestamp, 0).unwrap();
    datetime.humanize()
}

/// Calculate the number of files in the FileMetric result
// I'm lazy so this doesn't really need to be it's own function but it was easier to do this.
fn count_file_metrics(file_metrics: &Result<Vec<FileMetric>, Error>) -> usize {
    match file_metrics {
        Ok(metrics) => metrics.len(),
        Err(_) => 0,
    }
}

/// Guess the mimetype
/// This uses the magic library, which seems to be the most accurate out of the ones I've seen.
/// As the name suggests, this is magic. I have no idea HOW it works, but it should.
/// This doesn't need the full file loaded into memory, so it loads 2048 bytes.
fn detect_mime_type(file_data: &[u8]) -> Result<String, magic::MagicError> {
    COOKIE.with(|cookie| {
        let result = cookie.buffer(file_data)?;
        Ok(result)
    })
}

/// This spawns a tokio task to run a interval timer forever.
/// the interval timer runs every 'period' seconds.
pub fn cleaner_thread(period: i32) {
    let _forever = task::spawn(async move {
        let mut interval = time::interval(Duration::from_secs(period.try_into().unwrap()));
        let sqlconn = SQLITE.get().unwrap();
        loop {
            // Wait for the next interval
            interval.tick().await;
            // Get a vec of files that will expire this loop.
            // Also I'm sorta just realising how annoying it is to pass sqlconn through like 4 functions deep just to do something.
            let old_files_result = db::get_old_files(sqlconn).await;

            match old_files_result {
                Ok(file_list) => {
                    info!("[Cleaner] Running Cleaner");
                    for file in file_list {
                        // Delete the file from the database
                        db::delete_file(sqlconn, &file).await.unwrap_or_else(|err| {
                            error!(
                                "[Cleaner] Failed to delete file from database: {}, error: {:?}",
                                &file, err
                            );
                            0
                        });
                        // Delete the file from the filesystem
                        engine::delete_file(&file).await.unwrap_or_else(|err| {
                            error!("[Cleaner] Failed to delete file from database: {:?}", err);
                        });
                    }
                    info!("[Cleaner] Cleaner finished");
                }
                Err(err) => {
                    error!("[Cleaner] Error getting files to expire: {}", err);
                    // Return a empty Vec so it doesn't delete anything
                    return Vec::<String>::new();
                }
            }
        }
    });
}

/// This function checks if a file exists. Thats it.
/// We use this to detect files that don't exist in the db, but exist in reality.
async fn check_file_exists(filename: &str, directory: &str) -> bool {
    let full_path = format!("{}/{}", directory, filename);
    fs::metadata(full_path).is_ok()
}

/// This function is called when a file missing from the database /actually/ exists on the filesystem!
/// It adds it back to the database, with stub information.
pub async fn add_undeleted_file(filename: &str) -> Result<SqliteQueryResult, sqlx::Error> {
    // Grab a SQLITE pool
    let sqlconn = SQLITE
        .get()
        .expect("There was a problem grabbing a SQLite Pool");

    // Get the path for the file
    let path = PathBuf::from("files/").join(filename);

    // Guess the mimetype
    // Get the temporary file
    let mut mime_file = File::open(&path).expect("Error opening file");
    let mut buffer = vec![0; 2048];

    // Load up 2048 bytes (There wasn't a decent answer on how much or little I should read)
    mime_file
        .read_exact(&mut buffer)
        .expect("Error reading file");

    let mimetype = detect_mime_type(&buffer).unwrap_or("text/plain".to_string());

    // Check the filesize
    let filesize = fs::metadata(path).unwrap().len() as i32;

    // Calculate the expiry
    let expiry = calculate_expiry(sqlconn, filename, filesize).await;
    // LEGACY: This file shouldn't be able to be seen by anyone, since we don't really know the owner, and then nobody should be able to delete it.
    // Grab the first master key we can find.
    let masterkey = CONFIG.uploading.fallback_key.as_str();
    // Set accessed to now
    let accessed = SystemTime::now()
        .duration_since(SystemTime::UNIX_EPOCH)
        .unwrap();

    // Set 'domain' to "legacy"
    // since this is just for fancy stats, it'll work well when we don't know where it was uploaded.
    let domain = "legacy";

    // Add the file to the database
    db::add_file(
        sqlconn,
        filename,
        &mimetype,
        expiry,
        masterkey,
        accessed.as_secs() as i32,
        filesize,
        "127.0.0.1",
        domain,
    )
    .await
}

/// This function tries it's best at guessing if the user is a web browser or a API Request.
/// We do this buy checking if the form data exists from the upload request, but failing that, we check if the user agent is a browser.
/// If it is, we return true, otherwise false.
/// TODO: In the future, we'll move the api to a /api/... endpoint, and we won't need to check at all.
pub async fn is_browser(req: &mut Request) -> bool {
    // Unpack the form data, if it exists.
    let form_data = req.form_data().await;

    // We only need to check if the form exists for uploads, so this goes first.
    if let Ok(data) = form_data {
        // Form data exists, so this is at least a file upload
        // Check the form data itself for the "source" field.
        if data.fields.get("source").is_some() {
            debug!("Source detected from form, user is a web browser!");
            // source exists, so we know this is a web browser uploading a file.
            return true;
        }
    }

    // Check based on user-agent "Mozilla". Sorry to the obscure-ass web browsers I guess.
    if let Some(agent) = req.headers().get("user-agent") {
        if agent.to_str().unwrap_or("").contains("Mozilla") {
            debug!("User Agent says this is likely a web browser!");
            return true;
        } else {
            // We couldn't find "Mozilla" in the user-agent, so we'll assume it's an API user.
            debug!("User Agent says this is likely a API Request!");
            return false;
        }
    }
    // If somehow, there is NO user agent at all, just assume it's some idiot who removed it.
    debug!("We got nothing, assuming web browser!");
    true
}
