#[cfg(test)]
mod tests {
    use crate::handlers::{detect_mime_type, guess_ip, is_mimetype_banned};
    use salvo::conn::SocketAddr as SalvoSocketAddr;
    use salvo::hyper::HeaderMap;
    use std::net::{IpAddr, Ipv4Addr, Ipv6Addr, SocketAddr};

    // Test IP Guessing with no headers at all
    // This should end up guessing the ip from remote_addr in the request (sent through via 'ip' var)
    #[tokio::test]
    async fn test_guess_ip_no_headers() {
        let headers = HeaderMap::new();
        let ip_addr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1));
        let std_socket_addr = SocketAddr::new(ip_addr, 8080);
        let salvo_socket_addr = SalvoSocketAddr::from(std_socket_addr);
        let guessed_ip = guess_ip(&headers, &salvo_socket_addr);
        assert_eq!(guessed_ip, "127.0.0.1");
    }

    #[tokio::test]
    async fn test_guess_ip_cf_connecting_ip() {
        let mut headers = HeaderMap::new();
        headers.insert("CF-Connecting-IP", "192.168.0.1".parse().unwrap());
        let ip_addr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1));
        let std_socket_addr = SocketAddr::new(ip_addr, 8000);
        let salvo_socket_addr = SalvoSocketAddr::from(std_socket_addr);
        let guessed_ip = guess_ip(&headers, &salvo_socket_addr);
        assert_eq!(guessed_ip, "192.168.0.1");
    }

    #[tokio::test]
    async fn test_guess_ip_x_real_ip() {
        let mut headers = HeaderMap::new();
        headers.insert("X-Real-IP", "192.168.0.1".parse().unwrap());
        let ip_addr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1));
        let std_socket_addr = SocketAddr::new(ip_addr, 8000);
        let salvo_socket_addr = SalvoSocketAddr::from(std_socket_addr);
        let guessed_ip = guess_ip(&headers, &salvo_socket_addr);
        assert_eq!(guessed_ip, "192.168.0.1");
    }

    #[tokio::test]
    async fn test_guess_ip_ip_v4() {
        let headers = HeaderMap::new();
        let ip_addr = IpAddr::V4(Ipv4Addr::new(192, 168, 0, 1));
        let std_socket_addr = SocketAddr::new(ip_addr, 8000);
        let salvo_socket_addr = SalvoSocketAddr::from(std_socket_addr);
        let guessed_ip = guess_ip(&headers, &salvo_socket_addr);
        assert_eq!(guessed_ip, "192.168.0.1");
    }

    #[tokio::test]
    async fn test_guess_ip_ip_v6() {
        let headers = HeaderMap::new();
        let ip_addr = IpAddr::V6(Ipv6Addr::new(0x2001, 0xdb8, 0, 0, 0, 0, 0, 1));
        let std_socket_addr = SocketAddr::new(ip_addr, 8000);
        let salvo_socket_addr = SalvoSocketAddr::from(std_socket_addr);
        let guessed_ip = guess_ip(&headers, &salvo_socket_addr);
        assert_eq!(guessed_ip, "2001:db8::1");
    }

    // Testing banned files!

    // Mock implementation of `Config`
    struct MockConfig;
    impl MockConfig {
        fn get_array(&self, key: &str) -> Option<Vec<String>> {
            match key {
                "operations.banned_mimetype" => Some(vec![
                    String::from("application/x-msdownload"),
                    String::from("application/x-msi"),
                    String::from("application/x-silverlight"),
                ]),
                _ => None,
            }
        }
    }

    #[tokio::test]
    async fn test_is_mimetype_banned() {
        // Test when mimetype is banned
        let result = is_mimetype_banned("application/x-msdownload").await;
        assert!(result);

        // Test the 3rd mimetype in the list, hopefully to catch any errors with the loop
        let result = is_mimetype_banned("application/x-silverlight").await;
        assert!(result);

        // Test when mimetype is not banned
        let result = is_mimetype_banned("audio/mp3").await;
        assert!(!result);
    }

    #[tokio::test]
    async fn test_is_mimetype_banned_case_insensitive() {
        // Test when mimetype is banned (case-insensitive)
        let result = is_mimetype_banned("APPLICATION/x-msi").await;
        assert!(result);

        // Test when mimetype is not banned (case-insensitive)
        let result = is_mimetype_banned("aUdIo/MP4").await;
        assert!(!result);
    }

    #[tokio::test]
    async fn test_detect_mime_type() {
        // Test case 1: Valid file data
        let file_data = b"This is a test file";
        let result = detect_mime_type(file_data);
        println!("Test 1: Valid file data: {:?}", result);
        assert!(result.is_ok());
        assert_eq!(result.unwrap(), "text/plain");

        // Test case 2: Empty file data
        let file_data = b"";
        let result = detect_mime_type(file_data);
        println!("Test 2: Empty file data Result: {:?}", result);
        assert!(result.is_ok());
        assert_eq!(result.unwrap(), "application/x-empty");

        // Test case 3: small png file
        // This should test for a bug that lets really small files not count as the correct
        // filetype
        let small_png_data: &[u8] = &[
            // a random screenshot lol
            0x89, 0x50, 0x4e, 0x47, 0x0d, 0x0a, 0x1a, 0x0a, 0x00, 0x00, 0x00, 0x0d, 0x49, 0x48,
            0x44, 0x52, 0x00, 0x00, 0x00, 0x2f, 0x00, 0x00, 0x00, 0x20, 0x08, 0x06, 0x00, 0x00,
            0x00, 0x82, 0x71, 0x21, 0x79, 0x00, 0x00, 0x00, 0x01, 0x73, 0x52, 0x47, 0x42, 0x00,
            0xae, 0xce, 0x1c, 0xe9, 0x00, 0x00, 0x00, 0x04, 0x67, 0x41, 0x4d, 0x41, 0x00, 0x00,
            0xb1, 0x8f, 0x0b, 0xfc, 0x61, 0x05, 0x00, 0x00, 0x00, 0x09, 0x70, 0x48, 0x59, 0x73,
            0x00, 0x00, 0x0e, 0xc3, 0x00, 0x00, 0x0e, 0xc3, 0x01, 0xc7, 0x6f, 0xa8, 0x64, 0x00,
            0x00, 0x01, 0x23, 0x49, 0x44, 0x41, 0x54, 0x58, 0x47, 0xed, 0x97, 0xb1, 0xad, 0x84,
            0x30, 0x0c, 0x86, 0x33, 0x00, 0x4b, 0xb0, 0x03, 0x2b, 0xb0, 0x01, 0x0b, 0xd0, 0x43,
            0x4f, 0x4f, 0x49, 0x4b, 0x4b, 0x47, 0x4d, 0x4b, 0x4f, 0xcd, 0x4e, 0x39, 0x7d, 0x91,
            0x8c, 0xa2, 0x28, 0xb9, 0x3b, 0xbd, 0xc6, 0xef, 0x24, 0x17, 0x9f, 0x50, 0xfc, 0xc7,
            0xce, 0x1f, 0xc7, 0x0d, 0xae, 0xaa, 0x2a, 0xff, 0xab, 0x98, 0x79, 0x2d, 0xcc, 0xbc,
            0x16, 0x6e, 0x18, 0x06, 0xff, 0xab, 0x98, 0x79, 0x2d, 0xcc, 0xbc, 0x16, 0x66, 0x5e,
            0x0b, 0x33, 0xaf, 0x85, 0x99, 0xd7, 0xc2, 0xcc, 0x6b, 0x61, 0xe6, 0xb5, 0x30, 0xf3,
            0x5a, 0xd8, 0x6f, 0xa0, 0x16, 0x6a, 0xe6, 0xf7, 0x7d, 0xf7, 0xf7, 0x7d, 0x87, 0x6f,
            0xdb, 0xb6, 0xd9, 0x3d, 0x9f, 0x50, 0x33, 0x8f, 0xe1, 0xbe, 0xef, 0xfd, 0x71, 0x1c,
            0xfe, 0xba, 0xae, 0x3f, 0x5d, 0x40, 0x7d, 0x6c, 0xea, 0xba, 0xf6, 0xe7, 0x79, 0x86,
            0x17, 0xc8, 0xe9, 0xef, 0xf8, 0x17, 0x33, 0x3f, 0x4d, 0x53, 0x18, 0x21, 0x2e, 0x92,
            0xd3, 0x4b, 0x38, 0x92, 0x62, 0x44, 0xa0, 0xd0, 0xba, 0xae, 0x4f, 0x3c, 0x9d, 0x4d,
            0xd6, 0xcb, 0xb2, 0x3c, 0xb3, 0x4b, 0xf7, 0xba, 0xae, 0xfb, 0x3a, 0x3f, 0x86, 0x38,
            0x7b, 0xe2, 0x7c, 0x81, 0x3c, 0x6a, 0xa7, 0x71, 0x70, 0x24, 0x30, 0x7b, 0x24, 0x6f,
            0xdb, 0xf6, 0x08, 0x1c, 0xcc, 0x2c, 0x8e, 0xe3, 0x18, 0x74, 0x0a, 0x30, 0x9f, 0xa2,
            0x53, 0x54, 0x74, 0x6a, 0xa0, 0xc5, 0x87, 0x88, 0x4e, 0x6e, 0x4e, 0x4f, 0xe1, 0x7c,
            0x6a, 0xa5, 0xf1, 0xb7, 0xe6, 0xe9, 0x10, 0x85, 0x21, 0x7e, 0x36, 0x8a, 0xd1, 0x59,
            0x59, 0x53, 0x98, 0x98, 0x74, 0x8f, 0xa2, 0x5c, 0x50, 0xf4, 0x79, 0x9e, 0x83, 0x2e,
            0xeb, 0xd4, 0x8c, 0x74, 0xb7, 0xd4, 0x7d, 0x34, 0xc6, 0x27, 0xa7, 0x95, 0x70, 0x39,
            0xe3, 0x40, 0x31, 0x0c, 0xc9, 0x9a, 0xee, 0x11, 0x93, 0xa7, 0xc5, 0x7c, 0xac, 0xc7,
            0xe6, 0x65, 0x6f, 0x0e, 0xc9, 0x8f, 0x69, 0x9a, 0xa6, 0xa8, 0x95, 0xa9, 0xfc, 0x0b,
            0x8b, 0x4a, 0xde, 0xf6, 0xdf, 0xf9, 0xf2, 0xac, 0x00, 0x00, 0x00, 0x00, 0x49, 0x45,
            0x4e, 0x44, 0xae, 0x42,
        ];
        let result = detect_mime_type(small_png_data);
        println!("Test case 3: small_png: {:?}", result);
        assert!(result.is_ok());
        assert_eq!(result.unwrap(), "image/png");

        // Test case 4: putty.exe
        let file_data: &[u8] = &[
            // Putty.exe trimmed at 2048
            // This is stupid, and probably not needed for a test lol.
            0x4D, 0x5A, 0x78, 0x00, 0x01, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x78, 0x00, 0x00, 0x00, 0x0E, 0x1F, 0xBA, 0x0E, 0x00, 0xB4,
            0x09, 0xCD, 0x21, 0xB8, 0x01, 0x4C, 0xCD, 0x21, 0x54, 0x68, 0x69, 0x73, 0x20, 0x70,
            0x72, 0x6F, 0x67, 0x72, 0x61, 0x6D, 0x20, 0x63, 0x61, 0x6E, 0x6E, 0x6F, 0x74, 0x20,
            0x62, 0x65, 0x20, 0x72, 0x75, 0x6E, 0x20, 0x69, 0x6E, 0x20, 0x44, 0x4F, 0x53, 0x20,
            0x6D, 0x6F, 0x64, 0x65, 0x2E, 0x24, 0x00, 0x00, 0x50, 0x45, 0x00, 0x00, 0x64, 0x86,
            0x0A, 0x00, 0x91, 0x10, 0x5C, 0x63, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0xF0, 0x00, 0x22, 0x00, 0x0B, 0x02, 0x0E, 0x00, 0x00, 0x5A, 0x0E, 0x00, 0x00, 0x70,
            0x0A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x14, 0x88, 0x0B, 0x00, 0x00, 0x10, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x40, 0x01, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x02,
            0x00, 0x00, 0x06, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x60, 0x19, 0x00, 0x00, 0x04, 0x00, 0x00, 0x44, 0xA8,
            0x19, 0x00, 0x02, 0x00, 0x60, 0x81, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x90, 0x11,
            0x12, 0x00, 0xB4, 0x00, 0x00, 0x00, 0x00, 0xA0, 0x13, 0x00, 0x58, 0x9F, 0x05, 0x00,
            0x00, 0xD0, 0x12, 0x00, 0x78, 0x6C, 0x00, 0x00, 0x00, 0xCE, 0x18, 0x00, 0x28, 0x57,
            0x00, 0x00, 0x00, 0x40, 0x19, 0x00, 0xA8, 0x1E, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x0E, 0x10, 0x00, 0x28, 0x00, 0x00, 0x00,
            0xE0, 0xA6, 0x0F, 0x00, 0x40, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x28, 0x1D, 0x12, 0x00, 0xE0, 0x0A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x2E, 0x74, 0x65, 0x78, 0x74, 0x00, 0x00, 0x00,
            0x66, 0x59, 0x0E, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x5A, 0x0E, 0x00, 0x00, 0x04,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x20, 0x00, 0x00, 0x60, 0x2E, 0x72, 0x64, 0x61, 0x74, 0x61, 0x00, 0x00, 0xBC, 0xFF,
            0x03, 0x00, 0x00, 0x70, 0x0E, 0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x5E, 0x0E, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00,
            0x00, 0x40, 0x2E, 0x64, 0x61, 0x74, 0x61, 0x00, 0x00, 0x00, 0x9C, 0x55, 0x00, 0x00,
            0x00, 0x70, 0x12, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x5E, 0x12, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0xC0,
            0x2E, 0x70, 0x64, 0x61, 0x74, 0x61, 0x00, 0x00, 0x78, 0x6C, 0x00, 0x00, 0x00, 0xD0,
            0x12, 0x00, 0x00, 0x6E, 0x00, 0x00, 0x00, 0x6E, 0x12, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x40, 0x2E, 0x30,
            0x30, 0x63, 0x66, 0x67, 0x00, 0x00, 0x38, 0x00, 0x00, 0x00, 0x00, 0x40, 0x13, 0x00,
            0x00, 0x02, 0x00, 0x00, 0x00, 0xDC, 0x12, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x40, 0x2E, 0x67, 0x78, 0x66,
            0x67, 0x00, 0x00, 0x00, 0x60, 0x2A, 0x00, 0x00, 0x00, 0x50, 0x13, 0x00, 0x00, 0x2C,
            0x00, 0x00, 0x00, 0xDE, 0x12, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x40, 0x2E, 0x74, 0x6C, 0x73, 0x00, 0x00,
            0x00, 0x00, 0x11, 0x00, 0x00, 0x00, 0x00, 0x80, 0x13, 0x00, 0x00, 0x02, 0x00, 0x00,
            0x00, 0x0A, 0x13, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x40, 0x00, 0x00, 0xC0, 0x5F, 0x52, 0x44, 0x41, 0x54, 0x41, 0x00, 0x00,
            0x5C, 0x01, 0x00, 0x00, 0x00, 0x90, 0x13, 0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x0C,
            0x13, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x40, 0x00, 0x00, 0x40, 0x2E, 0x72, 0x73, 0x72, 0x63, 0x00, 0x00, 0x00, 0x58, 0x9F,
            0x05, 0x00, 0x00, 0xA0, 0x13, 0x00, 0x00, 0xA0, 0x05, 0x00, 0x00, 0x0E, 0x13, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00,
            0x00, 0x40, 0x2E, 0x72, 0x65, 0x6C, 0x6F, 0x63, 0x00, 0x00, 0xA8, 0x1E, 0x00, 0x00,
            0x00, 0x40, 0x19, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0xAE, 0x18, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x42,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x56, 0x57, 0x53, 0x48, 0x83, 0xEC, 0x40, 0x48, 0x89, 0xD6, 0x48, 0x89,
            0xCF, 0x4C, 0x89, 0x44, 0x24, 0x70, 0x4C, 0x89, 0x4C, 0x24, 0x78, 0x48, 0x8B, 0x05,
            0x32, 0x60, 0x12, 0x00, 0x48, 0x31, 0xE0, 0x48, 0x89, 0x44, 0x24, 0x38, 0x48, 0x8D,
            0x5C, 0x24, 0x70, 0x48, 0x89, 0x5C, 0x24, 0x30, 0xE8, 0xBB, 0x61, 0x00, 0x00, 0x48,
            0x8B, 0x08, 0x48, 0x83, 0xC9, 0x01, 0x48, 0x89, 0x5C, 0x24, 0x28, 0x48, 0xC7, 0x44,
            0x24, 0x20, 0x00, 0x00, 0x00, 0x00, 0x48, 0x89, 0xFA, 0x49, 0xC7, 0xC0, 0xFF, 0xFF,
            0xFF, 0xFF, 0x49, 0x89, 0xF1, 0xE8, 0x9C, 0x98, 0x0B, 0x00, 0x89, 0xC6, 0x85, 0xC0,
            0xB8, 0xFF, 0xFF, 0xFF, 0xFF, 0x0F, 0x48, 0xF0, 0x48, 0x8B, 0x4C, 0x24, 0x38, 0x48,
            0x31, 0xE1, 0xE8, 0x8B, 0x72, 0x0B, 0x00, 0x89, 0xF0, 0x48, 0x83, 0xC4, 0x40, 0x5B,
            0x5F, 0x5E, 0xC3, 0xCC, 0x56, 0x48, 0x83, 0xEC, 0x70, 0x48, 0x8B, 0x05, 0xC4, 0x5F,
            0x12, 0x00, 0x48, 0x31, 0xE0, 0x48, 0x89, 0x44, 0x24, 0x68, 0x48, 0x8B, 0x35, 0x6D,
            0x6E, 0x12, 0x00, 0x48, 0x85, 0xF6, 0x74, 0x20, 0x48, 0x83, 0x3D, 0x68, 0x6E, 0x12,
            0x00, 0x00, 0x74, 0x3A, 0x48, 0x8B, 0x4C, 0x24, 0x68, 0x48, 0x31, 0xE1, 0xE8, 0x49,
            0x72, 0x0B, 0x00, 0x48, 0x89, 0xF0, 0x48, 0x83, 0xC4, 0x70, 0x5E, 0xC3, 0x4C, 0x8B,
            0x05, 0x49, 0x61, 0x0E, 0x00, 0x31, 0xC9, 0x31, 0xD2, 0xE8, 0x60, 0x0F, 0x04, 0x00,
            0x48, 0x89, 0xC6, 0x48, 0x89, 0x05, 0x2E, 0x6E, 0x12, 0x00, 0x48, 0x83, 0x3D, 0x2E,
            0x6E, 0x12, 0x00, 0x00, 0x75, 0xC6, 0xC7, 0x44, 0x24, 0x20, 0x00, 0x00, 0x00, 0x00,
            0x48, 0x8D, 0x05, 0x6D, 0x00, 0x00, 0x00, 0x48, 0x89, 0x44, 0x24, 0x28, 0x48, 0xC7,
            0x44, 0x24, 0x30, 0x00, 0x00, 0x00, 0x00, 0x48, 0x8B, 0x0D, 0x10, 0x6E, 0x12, 0x00,
            0x48, 0x89, 0x4C, 0x24, 0x38, 0xBA, 0xC8, 0x00, 0x00, 0x00, 0xFF, 0x15, 0x20, 0x10,
            0x12, 0x00, 0x48, 0x89, 0x44, 0x24, 0x40, 0xBA, 0x01, 0x7F, 0x00, 0x00, 0x31, 0xC9,
            0xFF, 0x15, 0x06, 0x10, 0x12, 0x00, 0x48, 0x89, 0x44, 0x24, 0x48, 0x0F, 0x57, 0xC0,
            0x0F, 0x11, 0x44, 0x24, 0x50, 0x48, 0x8B, 0x05, 0xCA, 0x6D, 0x12, 0x00, 0x48, 0x89,
            0x44, 0x24, 0x60, 0x48, 0x8D, 0x4C, 0x24, 0x20, 0xFF, 0x15, 0x62, 0x10, 0x12, 0x00,
            0x48, 0x8B, 0x35, 0xB3, 0x6D, 0x12, 0x00, 0xE9, 0x50, 0xFF, 0xFF, 0xFF, 0xCC, 0xCC,
            0xCC, 0xCC, 0xCC, 0xCC, 0x41, 0x57, 0x41, 0x56, 0x41, 0x55, 0x41, 0x54, 0x56, 0x57,
            0x55, 0x53, 0x48, 0x81, 0xEC, 0x18, 0x09, 0x00, 0x00, 0x4D, 0x89, 0xCD, 0x4D, 0x89,
            0xC7, 0x89, 0xD6, 0x49, 0x89, 0xCE, 0x48, 0x8B, 0x05, 0xCB, 0x5E, 0x12, 0x00, 0x48,
            0x31, 0xE0, 0x48, 0x89, 0x84, 0x24, 0x10, 0x09, 0x00, 0x00, 0x81, 0xFA, 0xFF, 0x01,
            0x00, 0x00, 0x0F, 0x8F, 0x8A, 0x00, 0x00, 0x00, 0x8D, 0x86, 0x60, 0xFF, 0xFF, 0xFF,
            0x83, 0xF8, 0x77, 0x0F, 0x87, 0xC1, 0x01, 0x00, 0x00, 0x48, 0x8D, 0x0D, 0xC2, 0x39,
            0x00, 0x00, 0x48, 0x63, 0x04, 0x81, 0x48, 0x01, 0xC8, 0xFF, 0xE0, 0xB9, 0x03, 0x00,
            0x00, 0x00, 0x44, 0x89, 0xEA, 0xE8, 0x08, 0xEE, 0x02, 0x00, 0x4C, 0x89, 0xF8, 0x48,
            0x83, 0xE0, 0xFD, 0x48, 0x3D, 0xE5, 0x00, 0x00, 0x00, 0x0F, 0x85, 0xDF, 0x02, 0x00,
            0x00, 0x81, 0xFE, 0x00, 0x01, 0x00, 0x00, 0x0F, 0x85, 0x8C, 0x24, 0x00, 0x00, 0x4C,
            0x89, 0xB4, 0x24, 0x10, 0x01, 0x00, 0x00, 0xC7, 0x84, 0x24, 0x18, 0x01, 0x00, 0x00,
            0x00, 0x01, 0x00, 0x00, 0x4C, 0x89, 0xBC, 0x24, 0x20, 0x01, 0x00, 0x00, 0x41, 0x81,
            0xE5, 0xFF, 0xDF, 0x00, 0x00, 0x4C, 0x89, 0xAC, 0x24, 0x28, 0x01, 0x00, 0x00, 0x48,
            0x8D, 0x8C, 0x24, 0x10, 0x01, 0x00, 0x00, 0xFF, 0x15, 0x89, 0x10, 0x12, 0x00, 0x31,
            0xF6, 0xE9, 0x61, 0x24, 0x00, 0x00, 0x81, 0xFE, 0xDF, 0x02, 0x00, 0x00, 0x0F, 0x8F,
            0x91, 0x01, 0x00, 0x00, 0x8D, 0x86, 0x00, 0xFE, 0xFF, 0xFF, 0x83, 0xF8, 0x32, 0x0F,
            0x87, 0xBF, 0x03, 0x00, 0x00, 0x48, 0x8D, 0x0D, 0x0C, 0x3B, 0x00, 0x00, 0x48, 0x63,
            0x04, 0x81, 0x48, 0x01, 0xC8, 0xFF, 0xE0, 0x4C, 0x39, 0x3D, 0x88, 0x80, 0x12, 0x00,
            0x75, 0x15, 0x4C, 0x39, 0x2D, 0x87, 0x80, 0x12, 0x00, 0x75, 0x0C, 0x81, 0x3D, 0x57,
            0x80, 0x12, 0x00, 0x00, 0x02, 0x00, 0x00, 0x74, 0x1F, 0xB1, 0x01, 0xE8, 0x5A, 0x5E,
            0x00, 0x00, 0x4C, 0x89, 0x3D, 0x63, 0x80, 0x12, 0x00, 0x4C, 0x89, 0x2D, 0x64, 0x80,
            0x12, 0x00, 0xC7, 0x05, 0x36, 0x80, 0x12, 0x00, 0x00, 0x02, 0x00, 0x00, 0xB9, 0x05,
            0x00, 0x00, 0x00, 0x44, 0x89, 0xEA, 0xE8, 0x35, 0xED, 0x02, 0x00, 0x41, 0xF6, 0xC7,
            0x13, 0x0F, 0x84, 0x1F, 0x1E, 0x00, 0x00, 0xFF, 0x15, 0x5D, 0x0D, 0x12, 0x00, 0x4C,
            0x39, 0xF0, 0x0F, 0x85, 0x10, 0x1E, 0x00, 0x00, 0x31, 0xC0, 0x41, 0xF6, 0xC7, 0x10,
            0x0F, 0x94, 0xC0, 0x83, 0xC8, 0x02, 0x41, 0xF6, 0xC7, 0x01, 0xBE, 0x01, 0x00, 0x00,
            0x00, 0x0F, 0x44, 0xF0, 0xE8, 0x6F, 0xB3, 0x00, 0x00, 0x41, 0x89, 0xC6, 0x44, 0x89,
            0xFF, 0x83, 0xE7, 0x08, 0xC1, 0xEF, 0x03, 0x41, 0x83, 0xE7, 0x04, 0x41, 0xC1, 0xEF,
            0x02, 0x44, 0x89, 0xE9, 0xC1, 0xF9, 0x10, 0x8B, 0x2D, 0x4F, 0x77, 0x12, 0x00, 0x89,
            0xC8, 0x29, 0xE8, 0x83, 0xC0, 0x01, 0x45, 0x85, 0xED, 0x0F, 0x49, 0xC1, 0x2B, 0x05,
            0x94, 0x77, 0x12, 0x00, 0x99, 0xF7, 0xFD, 0x89, 0xC5, 0x44, 0x89, 0xE9, 0xC1, 0xE1,
            0x10, 0x41, 0x0F, 0xBF, 0xD5, 0x8B, 0x1D, 0x23, 0x77, 0x12, 0x00, 0x89, 0xD0, 0x29,
            0xD8, 0x83, 0xC0, 0x01, 0x85, 0xC9, 0x0F, 0x49, 0xC2, 0x2B, 0x05, 0x71, 0x77, 0x12,
            0x00, 0x99, 0xF7, 0xFB, 0x89, 0xC3, 0x89, 0xF1, 0xE8, 0x69, 0xB3, 0x00, 0x00, 0x48,
            0x8B, 0x0D, 0x4A, 0x77, 0x12, 0x00, 0x44, 0x88, 0x74, 0x24, 0x40, 0x40, 0x88, 0x7C,
            0x24, 0x38, 0x44, 0x88, 0x7C, 0x24, 0x30, 0x89, 0x6C, 0x24, 0x28, 0x89, 0x5C, 0x24,
            0x20, 0x89, 0xF2, 0x41, 0x89, 0xC0, 0x41, 0xB9, 0x04, 0x00, 0x00, 0x00, 0xE8, 0x7B,
            0x25, 0x01, 0x00, 0x31, 0xF6, 0xE9, 0x1B, 0x23, 0x00, 0x00, 0x8D, 0x46, 0xFF, 0x83,
            0xF8, 0x50, 0x0F, 0x87, 0x42, 0x0B, 0x00, 0x00, 0x48, 0x8D, 0x0D, 0xB1, 0x36, 0x00,
            0x00, 0x48, 0x63, 0x04, 0x81, 0x48, 0x01, 0xC8, 0xFF, 0xE0, 0x48, 0x8B, 0x0D, 0x99,
            0x6B, 0x12, 0x00, 0xBA, 0x89, 0x00, 0x00, 0x00, 0xE8, 0x37, 0xF6, 0x03, 0x00, 0x80,
            0x3D, 0x90, 0x5C, 0x12, 0x00, 0x00, 0x75, 0x0B, 0xB9, 0x01, 0x00, 0x00, 0x00, 0xFF,
            0x15, 0xD3, 0x0E, 0x12, 0x00, 0xC6, 0x05, 0x7C, 0x5C, 0x12, 0x00, 0x01, 0x31, 0xF6,
            0x31, 0xC9, 0xFF, 0x15, 0xE2, 0x0D, 0x12, 0x00, 0xE9, 0xC4, 0x22, 0x00, 0x00, 0x81,
            0xFE, 0x10, 0x03, 0x00, 0x00, 0x7F, 0x6B, 0x81, 0xFE, 0xE0, 0x02, 0x00, 0x00, 0x0F,
            0x84, 0xF5, 0x01, 0x00, 0x00, 0x81, 0xFE, 0x07, 0x03, 0x00, 0x00, 0x0F, 0x84, 0xC1,
            0x01, 0x00, 0x00, 0x81, 0xFE, 0x0F, 0x03, 0x00, 0x00, 0x0F, 0x85, 0xCB, 0x0A, 0x00,
            0x00, 0x48, 0x83, 0x3D, 0x09, 0x77, 0x12, 0x00, 0x00, 0x0F, 0x84, 0xC7, 0x1C, 0x00,
            0x00, 0xE8, 0x3E, 0x89,
        ];
        let result = detect_mime_type(file_data);
        println!("Test case 3: putty.exe: {:?}", result);
        assert!(result.is_ok());
        assert_eq!(
            result.unwrap(),
            "application/vnd.microsoft.portable-executable"
        );
    }
}
