use chrono::{TimeZone, Utc};
use salvo::{
    hyper::header::HOST, oapi::endpoint, prelude::StatusCode, writing::Text, Depot, Request,
    Response,
};
use std::{
    path::{Path, PathBuf},
    time::SystemTime,
};
use tokio::{
    fs::{self, File},
    io::AsyncReadExt,
};
use tracing::{debug, error, info, trace};

use super::guess_ip;
use crate::{
    db, engine,
    handlers::{
        detect_mime_type, is_browser, is_mimetype_banned, render_template, Generic, TemplateEnum,
    },
    keys::Key,
    CONFIG, SQLITE,
};

/// This endpoint handles file uploading.
///
/// This does a lot of checks to make sure the user isn't breaking any rules.
#[endpoint(tags("file-management"))]
pub async fn upload(req: &mut Request, res: &mut Response, depot: &mut Depot) {
    // Setup the Sqlite pool stuff for later
    let sqlconn = SQLITE.get().unwrap();
    // Clone out the header and remote_addr so we can guess the ip later
    let headers = &req.headers().clone();
    let remote_addr = &req.remote_addr().clone();
    // Get the host header for nicely setting up the response.
    let host = headers[HOST].to_str().unwrap_or("localhost:8282");
    debug!("upload(req): {:?}", req);

    // Check if the request was done in https or not by seeing if "x-forwarded-proto" exists.
    let mut https: String = "http".to_string();
    if req.headers().contains_key("x-forwarded-proto") {
        https = "https".to_string();
    }

    // Try to figure out if the user is on a web browser.
    let is_user_web = is_browser(req).await;

    // Grab the key from the auth handler.
    let key = depot
        .get::<Key>("client_key")
        .expect("Failed to grab API Key from depot")
        .clone();

    // This is the start of the file upload process
    // This handles both the web and json responses
    if let Some(file) = req.file("file").await {
        // Generate new filename.
        // Set up the filename length from the config
        let length = CONFIG.operations.filename_size as usize;
        let filename =
            engine::generate_filename(length, file.name().unwrap_or("file").to_string()).await;

        // Get the temporary file
        trace!("opening file for mimetype check: {:?}", file.path());
        let mime_file = File::open(file.path()).await.expect("Failed to open file");
        let mut buffer = Vec::with_capacity(512);

        // Load up 512 bytes (This breaks with a few filetypes, but meh. https://en.wikipedia.org/wiki/List_of_file_signatures)
        let _n = mime_file
            .take(512)
            .read_to_end(&mut buffer)
            .await
            .expect("Error reading file");
        trace!("read bytes for mime parsing: {:x?}", buffer);

        // Guess the mimetype
        let mimetype: &str = &detect_mime_type(&buffer).unwrap_or("text/plain".to_string());
        debug!("upload(mimetype): {:?}", mimetype);

        // Check mimetypes for normal keys only.
        // If keys are disabled, the key is initialised with 2 (normal) so it's always checked. (i hope)
        if key.key_level.as_i32() == 2 {
            info!(
                "[{:?} - {:?}] Validating mimetype for file '{}'",
                headers[HOST].to_str().unwrap(),
                guess_ip(headers, remote_addr),
                filename.to_str().unwrap_or("")
            );
            // Check if the mimetype is banned, and render the error pages.
            if is_mimetype_banned(mimetype).await {
                // Check if the upload came from the web
                if is_user_web {
                    let template_filename = "error.html";
                    let template = TemplateEnum::Generic(Generic {
                        domain: String::from(headers[HOST].to_str().unwrap()),
                        message1: String::from("Error 403"),
                        message2: String::from("That filetype is not allowed."),
                        ..Default::default()
                    });

                    let status_code = StatusCode::FORBIDDEN;

                    render_template(res, headers, template_filename, template, status_code).await;
                    return;
                } else {
                    // Otherwise, render the error in json for the API.
                    res.status_code(StatusCode::FORBIDDEN);
                    res.render(Text::Json(r#"{"error": "BlockedMimetype"}"#));
                    return;
                }
            }
        } else {
            info!(
                "[{:?} - {:?}] Bypassing Mimetype check for non-Normal-ranked user.",
                headers[HOST].to_str().unwrap(),
                guess_ip(headers, remote_addr)
            );
        }

        // Set Destination path.
        let dest: PathBuf = PathBuf::from("files/").join(&filename);
        // Copy the file
        if let Err(e) = std::fs::copy(file.path(), Path::new(&dest)) {
            // ... but if there was an error, complain about it.
            error!(
                "There was a problem uploading file: {:?}, error: {}",
                dest, e
            );
            // If web is true, render html, otherwise json.
            if is_user_web {
                let template_filename = "error.html";
                let template = TemplateEnum::Generic(Generic {
                    domain: String::from(headers[HOST].to_str().unwrap()),
                    message1: String::from("Error 500"),
                    message2: String::from(
                        "There was a problem uploading the file, please try again.",
                    ),
                    ..Default::default()
                });
                let status_code = StatusCode::INTERNAL_SERVER_ERROR;
                render_template(res, headers, template_filename, template, status_code).await;
                return;
            } else {
                res.status_code(StatusCode::INTERNAL_SERVER_ERROR);
                res.render(Text::Json(r#"{"error": "FileUploadServerError"}"#));
                return;
            }
        } else {
            // This branch is if the file upload worked properly. \o/

            // Calculate the expiry
            // file.path is the path to the temp file.
            // file.size seems to be always 'None', so we'll ask the filesystem.
            let path = Path::new(file.path());
            let filesize = fs::metadata(path).await.unwrap().len() as i32;
            debug!("upload(filesize): {:?}", filesize);
            let expiry =
                engine::calculate_expiry(sqlconn, &filename.to_str().unwrap(), filesize).await;
            // Convert that expiry to actual time and date (for display)
            let expiry_dt = Utc.timestamp_opt(expiry as i64, 0).unwrap();
            debug!("upload(expiry): {:?}", expiry);
            // Determine what ip type it is.
            debug!("upload(headers): {:?}", headers);
            let ip = guess_ip(headers, remote_addr);
            debug!("upload(ip): {:?}", ip);
            // Get the current time
            let accessed = SystemTime::now()
                .duration_since(SystemTime::UNIX_EPOCH)
                .unwrap()
                .as_secs() as i32;
            debug!("upload(accessed): {:?}", accessed);
            // Add file to the database, or print an error.
            if let Err(e) = db::add_file(
                sqlconn,
                filename.as_os_str().to_str().unwrap(), // wow
                mimetype,
                expiry,
                // expiry_override,
                &key.uuid, // The API Key from the user.
                accessed,
                filesize,
                ip.as_str(),
                host,
            )
            .await
            {
                error!("File upload error: {:?}", e);
            }
            info!(
                "[{:?}] New Upload: {:?}, owner: {}, ip: {:?}",
                headers[HOST].to_str().unwrap(),
                filename,
                &key.uuid,
                guess_ip(headers, remote_addr)
            );
            let fileurl = format!(
                "{}://{}/{}",
                https,
                String::from(headers[HOST].to_str().unwrap()),
                filename.to_string_lossy()
            );

            // Upload is done, lets show the stuff to the user!
            if is_user_web {
                let template_filename = "link.html";
                let template = TemplateEnum::Generic(Generic {
                    domain: String::from(headers[HOST].to_str().unwrap()),
                    fileurl: String::from(&fileurl),
                    adminurl: String::from(""),
                    message1: expiry_dt.to_string(),
                    message2: String::from(""),
                });
                let status_code = StatusCode::OK;
                render_template(res, headers, template_filename, template, status_code).await;
            } else {
                res.status_code(StatusCode::OK);
                res.render(Text::Json(format!(
                    r#"{{"file": "{}", "url": "{}", "expiry": {}}}"#,
                    &filename.to_string_lossy().to_string(),
                    fileurl,
                    expiry,
                )));
            }
        };
    } else if is_user_web {
        let template_filename = "error.html";
        let template = TemplateEnum::Generic(Generic {
            domain: String::from(headers[HOST].to_str().unwrap()),
            message1: String::from("Error 400"),
            message2: String::from("We couldn't read the file from the request."),
            ..Default::default()
        });

        let status_code = StatusCode::BAD_REQUEST;
        render_template(res, headers, template_filename, template, status_code).await;
        return;
    } else {
        res.status_code(StatusCode::BAD_REQUEST);
        res.render(Text::Json(r#"{"error": "BadRequest"}"#));
        return;
    };
}
