use salvo::Writer;
use std::str::FromStr;

use chrono::{DateTime, Utc};
use chrono_humanize::{Accuracy, HumanTime, Tense};
use salvo::{
    handler,
    hyper::header::HOST,
    oapi::{endpoint, extract::QueryParam},
    prelude::StatusCode,
    writing::Text,
    Depot, Request, Response,
};
use tracing::{debug, error, info};

use crate::{
    db,
    handlers::{
        convert_file_size, convert_unix_timestamp, count_file_metrics, render_template, Generic,
        TemplateEnum,
    },
    keys::{
        db_keys, {self, Key, KeyLevel},
    },
    SQLITE,
};

use super::Profile;

/// This handlers serves the user a nice page for managing uploaded files, generating new keys, and doing any other user actions
#[handler]
pub async fn user_profile(req: &mut Request, res: &mut Response, depot: &mut Depot) {
    let headers = req.headers().clone();
    // Setup db pool
    let sqlconn = SQLITE.get().unwrap();

    // Grab the key from the auth handler.
    let client_key = depot
        .get::<Key>("client_key")
        .expect("Failed to grab API Key from depot")
        .clone();

    // Humanise the creation date of the API Key.
    let key_creation = {
        let datetime: DateTime<Utc> =
            DateTime::from_timestamp(client_key.creation_time.into(), 0).unwrap();
        let human_time = HumanTime::from(datetime);
        human_time.to_text_en(Accuracy::Rough, Tense::Present)
    };

    // Key Generation and handling.

    // Get how many keys total, and available to generate.
    // If these fail, it's 0 out of 0 keys.
    let available_keys = keys::keys_utilisation(sqlconn, &client_key)
        .await
        .unwrap_or(0);
    let max_keys = keys::max_keys(sqlconn, &client_key).await.unwrap_or(0);

    debug!(
        "Key: {} has {} of {} key generations left.",
        client_key.uuid, available_keys, max_keys
    );

    // Render the key generation bar.
    let mut key_generator = String::new();

    // Calculate progress bar stuffs.
    let progress_percentage = (max_keys - available_keys) as f64 / max_keys as f64 * 100.0;
    let progress_class = if progress_percentage >= 75.0 {
        "critical"
    } else if progress_percentage >= 50.0 {
        "high"
    } else {
        "normal"
    };

    // Add the progress bar
    key_generator.push_str(&format!(
        r#"
        <div class='progress-container'>
            <span class='progress-text'>{}/{} keys used</span>
            <div class='progress-bar'>
                <div class='progress-bar-fill {}' style='width: {}%;'></div>
            </div>
        </div>
        "#,
        max_keys - available_keys,
        max_keys,
        progress_class,
        progress_percentage
    ));

    // If there's more than 0 available keys, show the generator for keys.
    if available_keys > 0 {
        debug!(
            "Key {} is allowed to generate keys, showing generator.",
            client_key.uuid
        );

        // First half of the generator html.
        let form_html_start = r#"
            <br>
            <form id="keygen-form">
                <textarea name="comment" id="comment" placeholder="Enter a comment"></textarea>
        "#;
        key_generator.push_str(form_html_start);

        // Second half of the generator html.
        let form_html_end = r#"
                <div class='keygen-submit'>
                    <button type="button" onclick="submitForm()">Generate Key</button>
                </div>
            </form>
            <div id="response-message"></div>
        "#;

        // Only add the dropdown when the user is a master key.
        if client_key.key_level == KeyLevel::Master {
            debug!("Key is master.");
            let key_level_dropdown = r#"
                <div class="keygen-level">
                    <select name="keygen-level" id="keygen-level">
                        <option value="Normal">Normal</option>
                        <option value="Super">Super</option>
                    </select>
                </div>
            "#;
            key_generator.push_str(key_level_dropdown);
        }

        // Append the rest of the form after the optional dropdown.
        key_generator.push_str(form_html_end);
    }

    // Get a list of all children keys.
    let child_uuid = client_key.child_uuids.clone();

    let mut key_list_html = String::new();

    if child_uuid.is_empty() {
        debug!("No Keys, pushing message");
        key_list_html.push_str("<p>You don't have any generated Keys....yet.</p>");
    } else {
        // Add the table header.
        debug!("We found some keys, lets add the table!");
        key_list_html.push_str(
            r#"
            <table class="sortable">
                <thead><tr><th>Key</th><th>Creation Time</th><th>Comment</th></tr></thead>
                <tbody>
        "#,
        );
        // Loop through all keys, adding them into a table.
        for child_uuid in child_uuid {
            debug!("Starting loop for key {}", child_uuid);
            // Error handle the fetching of child keys
            match db_keys::get_key(sqlconn, &child_uuid).await {
                Ok(key) => {
                    // Humanise the creation_time
                    let human_creation_time = {
                        let datetime: DateTime<Utc> =
                            DateTime::from_timestamp(key.creation_time.into(), 0).unwrap();
                        let human_time = HumanTime::from(datetime);
                        human_time.to_text_en(Accuracy::Rough, Tense::Past)
                    };
                    key_list_html.push_str(&format!(
                        "<tr><td data-sort='{}'>{}</td><td data-sort='{}'>{}</td><td>{}</td></tr>",
                        key.uuid,
                        key.uuid,
                        key.creation_time,
                        human_creation_time,
                        key.comment.unwrap_or("".to_string()),
                    ));
                }
                Err(err) => {
                    error!(
                        "Failed to fetch child key (UUID: {}): {:?}",
                        child_uuid, err
                    );
                    key_list_html.push_str(&format!(
                        "<tr><td data-sort='{}'>{}</td><td data-sort='ERR'>Database Error</td><td>Database Error</td></tr>", child_uuid, child_uuid));
                    continue; // Skip to the next child UUID if an error occurs
                }
            };
        }
        // Add the table footer. (yeah that's it lol)
        key_list_html.push_str("</tbody></table>");
        debug!("key_list: {}", key_list_html);
    }

    // File Listing and handling.
    // This shows no matter the key generation part.

    // This returns a Vec of all the files uploaded by the given user key.
    let files = db::get_files_by_key(sqlconn, &client_key).await;

    // Generate HTML to list all the files in a table.
    let mut file_list_html = String::new();
    // For each file in the list, add the html to the rendered string
    match files.as_ref() {
        Ok(file_metrics) => {
            for f in file_metrics {
                // Change the filename to a file url
                let fileurl = format!("/{}", &f.filename);

                // Change the filesize to a human readable value
                let human_size = convert_file_size(f.filesize);

                // Change the expiry to a human readable value
                let human_time = convert_unix_timestamp(f.expiry);

                // Generate a deletion URL unique to the file.
                let deletion_url = format!("/api/v1/file/{}", &f.filename);

                // If is_deleted, change colour to red, else green
                let colour = if f.is_deleted {
                    "background-color: #ff00000a;".to_string()
                } else {
                    "background-color: #00ff100a;".to_string()
                };

                // Add a new file to the rendered string
                file_list_html.push_str(&format!(
                    "<tr style='{}'><td><a href={}>{}</a></td><td data-sort='{}'>{}</td><td>{}</td><td>{}</td><td data-sort='{}'>{}</td><td><button class=\"delete-button\" onclick=\"changeRowColor(this, '{}')\">Delete</button></td></tr>",
                    colour,
                    fileurl.as_str(),
                    f.filename.as_str(),
                    f.filesize,
                    human_size.as_str(),
                    f.mimetype.as_str(),
                    f.views.to_owned(),
                    f.expiry.to_owned(),
                    human_time.to_owned(),
                    deletion_url.to_owned()
                ));
            }
        }
        Err(err) => {
            // Log error to console
            error!("Error retrieving file metrics: {:?}", err);

            // Render a nice error page for the poor user
            let template_filename = "error.html";
            let template = TemplateEnum::Generic(Generic {
                domain: String::from(headers[HOST].to_str().unwrap_or_default()),
                message1: String::from("Error 500: Internal Server Error"),
                message2: String::from(
                    "Something went wrong while reading the result from the database, uh oh :(",
                ),
                ..Generic::default() // Use default to fill in remaining fields
            });

            let status_code = StatusCode::INTERNAL_SERVER_ERROR;
            render_template(res, &headers, template_filename, template, status_code).await;
        }
    }

    // Close the table when all files are 'rendered'
    file_list_html.push_str("</table>");

    // Log this to the server console
    info!(
        "[{}] New Request: /user_profile ({} Files)",
        &headers[HOST].to_str().unwrap(),
        count_file_metrics(&files)
    );

    let template_filename = "user_profile.html";
    let template = TemplateEnum::Profile(Profile {
        domain: String::from(headers[HOST].to_str().unwrap()),
        key_creation,
        key_generator,
        key_list_html,
        file_list_html,
    });
    let status_code = StatusCode::OK;

    render_template(res, &headers, template_filename, template, status_code).await
}

/// This handlers works as the API for generating new keys.
/// It handles the POST request for making a new key.
#[endpoint(
    tags("user-management"),
    parameters(
        ("comment", description = "The user comment for the newly created key."),
        ("key_level", description = "The permission level of the key. (Normal or Super)"),
    )
)]
pub async fn generate_key(
    req: &mut Request,
    res: &mut Response,
    depot: &mut Depot,
    key_level: QueryParam<String, false>,
    comment: QueryParam<String, false>,
) {
    let headers = req.headers().clone();
    // Setup db pool
    let sqlconn = SQLITE.get().unwrap();

    // Grab the key from the auth handler.
    let client_key = depot
        .get::<Key>("client_key")
        .expect("Failed to grab API Key from depot")
        .clone();

    // Turns out we haven't checked if the user is allowed to generate keys.
    // This lets the user to create as many keys as they want from the endpoint. oops!
    let available_keys = keys::keys_utilisation(sqlconn, &client_key)
        .await
        .unwrap_or(0);
    debug!("Available keys {}", available_keys);
    if available_keys <= 0 {
        // User shouldn't be able to generate any more keys.
        res.status_code(StatusCode::FORBIDDEN);
        res.render(Text::Json(r#"{"error": "NoAvailableKeys"}"#));
        return;
    }

    // Bug: this sucks since the QueryParam returns some odd option type, lets get the actual option inside.
    let comment = comment.into_inner();
    let key_level = key_level.into_inner();

    debug!("key_level to create: {:?}", key_level);
    debug!("comment: {:?}", comment);

    let key_level = KeyLevel::from_str(&key_level.unwrap_or("Normal".to_string()));
    // Unpack key level
    match key_level {
        Ok(key_level) => {
            // Check if the key is actually allowed to generate superkey.
            // If the client key isn't master, but the requested key level to generate is super, fail.
            if client_key.key_level != KeyLevel::Master && key_level == KeyLevel::Super {
                // User isn't allowed to generate super keys. We should complain.
                res.status_code(StatusCode::FORBIDDEN);
                res.render(Text::Json(r#"{"error": "NoPermissionForSuperKey"}"#));
                return;
            }
            // Generate key
            let new_key = db_keys::create_key(sqlconn, key_level, Some(client_key), comment).await;
            match new_key {
                Ok(new_key) => {
                    info!(
                        "[{}] Created new {} key - {}",
                        headers[HOST].to_str().unwrap(),
                        new_key.key_level,
                        new_key.uuid
                    );
                    // Generate a response saying that we did it!
                    res.status_code(StatusCode::OK);
                    let render_json = serde_json::json!({
                        "success": "Created Key",
                        "key": new_key.uuid,
                    })
                    .to_string();
                    res.render(Text::Json(render_json));
                }
                Err(e) => {
                    error!("Error while generating new key: {}", e);
                    res.status_code(StatusCode::INTERNAL_SERVER_ERROR);
                    res.render(Text::Json(r#"{"error": "UnknownErrorCreatingKey"}"#));
                    return;
                }
            }
        }
        Err(_) => {
            res.status_code(StatusCode::FORBIDDEN);
            res.render(Text::Json(r#"{"error": "NoPermissionForSuperKey"}"#));
            return;
        }
    };
}
