use serde_json::from_str;
use sqlx::{Pool, Sqlite};
use std::time::SystemTime;
use thiserror::Error;
use tracing::{debug, error, info, warn};

use crate::CONFIG;

use super::{generate_key, Key, KeyLevel};

// Enum for Key Checking.
#[derive(Error, Debug)]
pub enum KeyCheckError {
    #[error("Key not found")]
    NotFound,
    #[error("Key revoked: {0}")]
    Revoked(String),
    #[error("Database error: {0}")]
    DatabaseError(#[from] sqlx::Error),
}

#[derive(Error, Debug)]
pub enum KeyError {
    #[error("Key not found")]
    NotFound,
    #[error("Database error: {0}")]
    DatabaseError(#[from] sqlx::Error),
}

/// Initialise a new key for a new database.
/// This should only run once per server (unless you mess with the db)
pub async fn init_key(sqlconn: &Pool<Sqlite>) {
    // Count the total number of keys in the database.
    let key_count = sqlx::query!(
        r#"
            SELECT COUNT(*) as count FROM keys
        "#
    )
    .fetch_one(sqlconn)
    .await;

    if key_count
        .expect("Failed to check the key count from the db")
        .count
        == 0
    {
        warn!("Couldn't find any keys in the database, making a new master key now!");
        let gen_key = create_key(
            sqlconn,
            KeyLevel::Master,
            None,
            Some("Initial Master Key".to_string()),
        )
        .await;
        let gen_key = gen_key.expect("Failed to generate a new mater key... oops!");
        info!("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
        info!("Generated New Master Key: {}", gen_key.uuid);
        info!("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
    }
}

/// This adds a generated key into the DB.
/// We don't check any key validation here, this is done before calling.
pub async fn create_key(
    sqlconn: &Pool<Sqlite>,
    level: KeyLevel,
    parent_key: Option<Key>,
    comment: Option<String>,
) -> Result<Key, sqlx::Error> {
    // Ensure parent key is present for non-master keys
    if parent_key.is_none() && level != KeyLevel::Master {
        return Err(sqlx::Error::Configuration(
            "Parent key is missing".to_string().into(),
        ));
    }

    // Generate a new key
    let uuid = generate_key(sqlconn).await.unwrap();
    // Check if the comment is present, and add it if so.
    let comment_value = comment.as_deref();
    // Convert the KeyLevel to a i32.
    let keylevel = level.as_i32();
    // Convert `Option<Key>` to `Option<String>`
    let parent_key = parent_key.map(|key| key.uuid);
    // Generate the current unix time for key creation.
    let creation_time = SystemTime::now()
        .duration_since(SystemTime::UNIX_EPOCH)
        .unwrap()
        .as_secs() as i32;

    // Insert the key into the DB
    let _result = sqlx::query!(
        "INSERT INTO keys (
                uuid,
                key_level,
                parent_key,
                creation_time,
                comment)
                VALUES ( ?,?,?,?,? )",
        uuid,
        keylevel,
        parent_key,
        creation_time,
        comment_value,
    )
    .execute(sqlconn)
    .await?;

    // Construct the key.
    let final_key = Key {
        uuid,
        key_level: level,
        parent_key,
        creation_time,
        revoked: None,
        comment,
        child_uuids: vec![], // New keys can't have kids!
    };

    // Return the Key
    Ok(final_key)
}

/// This returns a key when given a client key.
/// It can also return a few errors.
/// If allow_unverified_keys is true, we immediately return a Fake Key result, since we don't need to validate anything.
pub async fn get_key(sqlconn: &Pool<Sqlite>, client_key: &str) -> Result<Key, KeyError> {
    if CONFIG.uploading.allow_unverified_keys {
        debug!("Ephemeral is not configured to do key validation, returning 'fake' Key.");
        return Ok(Key::new(
            client_key.to_string(),
            KeyLevel::Normal,
            None,
            None,
            0_i32,
            None,
            vec![],
        ));
    }

    let row = sqlx::query!(
        r#"
            SELECT 
                uuid, 
                key_level, 
                parent_key, 
                revoked, 
                creation_time, 
                comment, 
                child_uuids 
            FROM keys_with_children 
            WHERE uuid = ?
        "#,
        client_key
    )
    .fetch_one(sqlconn)
    .await;

    match row {
        Ok(row) => {
            // Convert the key_level from i32 to KeyLevel
            let key_level = KeyLevel::from_i32(
                row.key_level
                    .try_into()
                    .expect("Failed to convert key from i64 to i32"),
            )
            .ok_or_else(|| {
                KeyError::DatabaseError(sqlx::Error::Decode("Invalid key level".into()))
            })?;

            // Convert child_uuids from JSON string to Vec<String>
            let child_uuids: Vec<String> = from_str(&row.child_uuids.unwrap_or("[]".to_owned()))
                .unwrap_or_else(|_| {
                    debug!("child_uuids is empty or invalid, returning an empty Vec<String>");
                    vec![]
                });
            debug!("Conversion to Vec<Str>: {:#?}", child_uuids);

            let key = Key::new(
                row.uuid,
                key_level,
                row.parent_key,
                row.revoked,
                row.creation_time as i32,
                row.comment,
                child_uuids,
            );
            Ok(key)
        }
        Err(e) => {
            error!("Couldn't find key: {}", e);
            Err(KeyError::NotFound)
        }
    }
}

/// This checks if a key already exists in the table.
/// Used to avoid collisions (not that I can see many happening)
pub async fn check_key(sqlconn: &Pool<Sqlite>, key: &str) -> Result<(), KeyCheckError> {
    // Select the uuid and revoked columns from the table.
    let row = sqlx::query!("SELECT uuid, revoked FROM keys WHERE uuid = ?", key)
        .fetch_optional(sqlconn)
        .await?;

    match row {
        Some(record) => {
            if let Some(revoked_reason) = record.revoked {
                Err(KeyCheckError::Revoked(revoked_reason))
            } else {
                Ok(())
            }
        }
        None => Err(KeyCheckError::NotFound),
    }
}

/// Count the total number of keys in the database.
pub async fn count_keys(sqlconn: &Pool<Sqlite>) -> Result<i32, sqlx::Error> {
    // Count the total number of keys in the database.
    let key_count = sqlx::query!(
        r#"
            SELECT COUNT(*) as count FROM keys
        "#
    )
    .fetch_one(sqlconn)
    .await?;

    Ok(key_count.count)
}

/// This function retrieves every single key from the database.
/// This is used for viewing the full 'key tree' on the admin page.
pub async fn get_all_keys(sqlconn: &Pool<Sqlite>) -> Result<Vec<Key>, KeyError> {
    let results = sqlx::query!(
        r#"
            SELECT 
                uuid, 
                key_level, 
                parent_key, 
                revoked, 
                creation_time, 
                comment, 
                child_uuids 
            FROM keys_with_children 
        "#,
    )
    .fetch_all(sqlconn)
    .await?;

    // Initialise the empty vec
    let mut keys: Vec<Key> = vec![];

    // Iterate over all keys returned.
    for row in results {
        // Convert all key_level(i32) to KeyLevel
        let key_level = KeyLevel::from_i32(
            row.key_level
                .try_into()
                .expect("Failed to convert key from i64 to i32"),
        )
        .ok_or_else(|| KeyError::DatabaseError(sqlx::Error::Decode("Invalid key level".into())))?;

        // Convert child_uuids from JSON string to Vec<String>
        let child_uuids: Vec<String> = from_str(&row.child_uuids.unwrap_or("[]".to_owned()))
            .unwrap_or_else(|_| {
                debug!("child_uuids is empty or invalid, returning an empty Vec<String>");
                vec![]
            });

        // Create a new key
        let key = Key::new(
            row.uuid,
            key_level,
            row.parent_key,
            row.revoked,
            row.creation_time as i32,
            row.comment,
            child_uuids,
        );
        // Push the new key to the vec
        keys.push(key)
    }
    // Now that we're done, return the Vector
    Ok(keys)
}

/// This revokes a key.
/// Requires a revoke message.
/// Returns a bool if it worked.
pub async fn revoke_key(
    sqlconn: &Pool<Sqlite>,
    key: &str,
    revoke_message: &str,
) -> Result<u64, sqlx::Error> {
    // Update the revoked message for the provided key.
    let result = sqlx::query!(
        "UPDATE keys SET revoked = ? WHERE uuid = ?",
        revoke_message,
        key
    )
    .execute(sqlconn)
    .await?;

    let rows_changed = result.rows_affected();

    Ok(rows_changed)
}

/// This revokes a key.
/// Requires a revoke message.
/// Returns a bool if it worked.
pub async fn edit_comment(
    sqlconn: &Pool<Sqlite>,
    uuid: &str,
    comment: &str,
) -> Result<bool, sqlx::Error> {
    // Update the revoked message for the provided key.
    let _result = sqlx::query!("UPDATE keys SET comment = ? WHERE uuid = ?", comment, uuid)
        .execute(sqlconn)
        .await?;

    Ok(true)
}
