// Submodules for each request handler
pub mod db_keys;

// This page contains the backend logic for generating new keys
use core::fmt;
use std::str::FromStr;

use db_keys::KeyError;
use nanoid::nanoid;
use sqlx::{Pool, Sqlite};
use tracing::{debug, error};

use crate::{
    keys::db_keys::{check_key, get_key, KeyCheckError},
    CONFIG,
};

/// Key Object
/// The key contains basically everything you need to handle keys within the system.
/// It stores the users permissions, the creator, and a few optionals.
#[derive(Clone, Debug)]
pub struct Key {
    pub uuid: String,               // Actual Key
    pub key_level: KeyLevel,        // Permission Level
    pub parent_key: Option<String>, // Optional (since master keys) Parent Key.
    pub revoked: Option<String>,    // Optional Ban Reason.
    pub creation_time: i32,         // Key Creation Time.
    pub comment: Option<String>,    // Optional Key Name.
    pub child_uuids: Vec<String>,   // Vector of children.
}
impl Key {
    pub fn new(
        uuid: String,
        key_level: KeyLevel,
        parent_key: Option<String>,
        revoked: Option<String>,
        creation_time: i32,
        comment: Option<String>,
        child_uuids: Vec<String>,
    ) -> Self {
        Key {
            uuid,
            key_level,
            parent_key,
            revoked,
            creation_time,
            comment,
            child_uuids,
        }
    }
}

/// Possible Permission Levels a key can be.
#[derive(Debug, PartialEq, Clone)]
pub enum KeyLevel {
    Master,
    Super,
    Normal,
}

impl KeyLevel {
    pub fn as_i32(&self) -> i32 {
        match self {
            KeyLevel::Master => 0,
            KeyLevel::Super => 1,
            KeyLevel::Normal => 2,
        }
    }

    pub fn from_i32(val: i32) -> Option<KeyLevel> {
        match val {
            0 => Some(KeyLevel::Master),
            1 => Some(KeyLevel::Super),
            2 => Some(KeyLevel::Normal),
            _ => None,
        }
    }

    pub fn as_str(&self) -> &'static str {
        match self {
            KeyLevel::Master => "Master",
            KeyLevel::Super => "Super",
            KeyLevel::Normal => "Normal",
        }
    }
}

impl FromStr for KeyLevel {
    type Err = ();

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "Master" => Ok(KeyLevel::Master),
            "Super" => Ok(KeyLevel::Super),
            "Normal" => Ok(KeyLevel::Normal),
            _ => Err(()),
        }
    }
}

impl fmt::Display for KeyLevel {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.as_str())
    }
}

/// Generate a new Key
/// This should check if the key exists for another file, and regenerate if it does.
/// This does NOT give it a permission level, this literally just generates the key itself.
pub async fn generate_key(sqlconn: &Pool<Sqlite>) -> Result<String, KeyCheckError> {
    // Get the size of the apikey from the config
    let key_size: usize = CONFIG.operations.apikey_size.into();
    // Maximum attempts we'll try to generate a valid key.
    let max_attempts = 8;

    // Loop until we get a valid key
    for _ in 0..max_attempts {
        let key = nanoid!(key_size);

        match check_key(sqlconn, &key).await {
            Ok(_) => {
                debug!("generate_key: Found Existing Key, trying again.");
            }
            Err(KeyCheckError::NotFound) => {
                debug!("generate_key: Generated new key: {:?}", key);
                return Ok(key);
            }
            Err(KeyCheckError::Revoked(reason)) => {
                // Log revoked keys for debugging purposes
                debug!("generate_key: Found revoked key: {}", reason);
                // Continue to the next iteration to generate a new key
            }
            Err(KeyCheckError::DatabaseError(e)) => {
                // Log the database error
                debug!("generate_key: Database error: {}", e);
                return Err(KeyCheckError::DatabaseError(e));
            }
        }
    }
    // If the whole loop runs, and we don't return a valid key, we've hit the max attempts.
    error!("generate_key: Exceeded maximum key generations, generating a super long key.");
    // Generate a REALLY big id. If we ever see one of these, we're in trouble.
    Ok(nanoid!(32))
}

/// This function checks if key has been fully utilised.
/// It returns an i32, which shows how many keys generations are left. 0 being fully utilised.
pub async fn keys_utilisation(sqlconn: &Pool<Sqlite>, key: &Key) -> Result<i32, KeyError> {
    // If we fail to get the number of children, we say it's the max and we can't do any.
    let children_count: i32 = key.child_uuids.len().try_into().unwrap_or(i32::MAX);
    debug!("[{}] Children count: {}", key.uuid, children_count);

    let max_allowed_keys = max_keys(sqlconn, key).await?;

    // I wonder what would happen if children count is larger?
    let remaining_keys = max_allowed_keys - children_count;
    debug!("[{}] remaining_keys: {}", key.uuid, remaining_keys);

    Ok(remaining_keys)
}

/// This calculates how many keys a key is allowed to generate.
/// The below function uses this to show how many keys are left to generate.
pub async fn max_keys(sqlconn: &Pool<Sqlite>, key: &Key) -> Result<i32, KeyError> {
    match key.key_level {
        KeyLevel::Master => {
            // Master keys have 'infinite' key gens.
            Ok(i32::MAX)
        }
        KeyLevel::Super => {
            // Super keys have the maximum configured key gens.
            Ok(CONFIG.uploading.max_key_generations)
        }
        KeyLevel::Normal => {
            // Normal keys are halved for each step it takes to get to a Super/Master Key
            let steps = count_to_super(sqlconn, key.clone()).await?;
            let max =
                CONFIG.uploading.max_key_generations / 2_i32.pow(steps.try_into().unwrap_or(0));
            Ok(max)
        }
    }
}

/// Count how many parent keys it takes to get to a Super key.
/// This code is only run for Normal Keys.
pub async fn count_to_super(sqlconn: &Pool<Sqlite>, key: Key) -> Result<i32, KeyError> {
    // Initialize the current key and count to super.
    let mut current_key = key;
    let mut count = 0;

    while current_key.key_level == KeyLevel::Normal {
        // Increment the count
        count += 1;

        // Get the parent key
        if let Some(parent_uuid) = current_key.parent_key.as_ref() {
            match get_key(sqlconn, parent_uuid).await {
                Ok(parent_key) => match parent_key.key_level {
                    KeyLevel::Master => {
                        debug!("Root Parent is Master!");
                        break;
                    }
                    KeyLevel::Super => {
                        debug!("Found the root Super Key!");
                        break;
                    }
                    KeyLevel::Normal => {
                        debug!("We haven't found the root key, lets keep going!");
                        current_key = parent_key;
                    }
                },
                Err(e) => {
                    // Failed to get key - passing up.
                    return Err(e);
                }
            }
        } else {
            // If there is no parent key, return an error
            return Err(KeyError::NotFound);
        }
    }
    Ok(count)
}
