use figment::{
    providers::{Env, Format, Toml},
    Figment,
};
use magic::Cookie;
use once_cell::sync::OnceCell;
use salvo::prelude::*;
use salvo::serve_static::{StaticDir, StaticFile};
use sqlx::SqlitePool;

use lazy_static::lazy_static;
use std::fs::create_dir_all;
use tracing::{error, info};
use tracing_subscriber::filter::EnvFilter;
use tracing_subscriber::fmt;
use tracing_subscriber::prelude::*;

use crate::config::AppConfig;
use crate::db::DatabaseType;
use crate::handlers::auth::require_auth;
use crate::handlers::cleaner_thread;

mod config;
mod db;
mod engine;
mod handlers;
mod keys;

mod db_tests;
#[cfg(test)]
mod engine_tests;

// I /don't think/ there's any harm to adding multiple of thse for each sql backend.
static SQLITE: OnceCell<SqlitePool> = OnceCell::new();

// Initialise the magic db
// This is thread-local, so each thread will have it's own database loaded.
// It's probably not the best way of doing it
// and we could probably have a dedicated thread for mimetypes but idk i'm not that smart
thread_local! {
    static COOKIE: Cookie = {
        let cookie = Cookie::open(magic::CookieFlags::MIME_TYPE).unwrap();
        cookie.load::<&str>(&[]).unwrap();
        cookie
    };
}

// Setup the config globally because I can't figure out how to pass it to functions I don't call directly.
// This is evaluated at runtime, and not compilation. \o/
// Also, we try to merge configs. so envvars > config > defaults > error
lazy_static! {
    pub static ref CONFIG: AppConfig = {
        let config = Figment::new()
            .merge(Toml::file("./data/config.toml"))
            .merge(Env::prefixed("EPH_"))
            .extract()
            .unwrap_or_default();
        tracing::info!("Loaded Config!");
        tracing::debug!("full config: {:?}", config);
        config
    };
}

#[tokio::main]
async fn main() {
    // Initialise Logging
    let log_level = &CONFIG.logging.level;
    tracing_subscriber::registry()
        .with(fmt::layer().compact())
        .with(EnvFilter::new(log_level))
        .init();

    let sql_backend = &CONFIG.database.sql_backend;
    let db_url = &CONFIG.database.url;

    match DatabaseType::from_str(sql_backend) {
        Some(DatabaseType::Sqlite) => {
            info!("DB: Using SQLite backend `{}`", &db_url);
            // TODO: Check if DB Exists, if not, create.

            let pool = SqlitePool::connect(db_url).await.unwrap();
            SQLITE.set(pool).unwrap();
        }
        Some(DatabaseType::Postgres) => {
            info!("DB: Using Postgres backend");
            error!("https://github.com/launchbadge/sqlx/issues/964");
            todo!("Postgres not implemented yet")
            // let pool = PgPool::connect(&psql).await;
            // TODO: figure out how to use sqlx::Pool struct https://docs.rs/sqlx/latest/sqlx/struct.Pool.html
        }
        Some(DatabaseType::Mysql) => {
            info!("DB: Using MySQL backend");
            error!("https://github.com/launchbadge/sqlx/issues/964");
            todo!("MySQL not implemented yet")
        }
        Some(DatabaseType::Oraclecloud) => {
            info!("DB: Using OracleCloud backend");
            error!("https://github.com/launchbadge/sqlx/issues/964");
            todo!("OracleCloud not implemented yet")
        }
        _ => panic!("Unsupported SQL backend: {}", sql_backend),
    };

    // There's probably a better way, but this is where we generate a new master key.
    keys::db_keys::init_key(SQLITE.get().unwrap()).await;

    info!(
        "Starting cleaner thread every {} seconds",
        CONFIG.operations.cleaner_interval
    );
    cleaner_thread(CONFIG.operations.cleaner_interval);

    // Attempt to create the files directory
    if let Err(e) = create_dir_all("files") {
        error!("Couldn't create files directory: {}", e);
        std::process::exit(1);
    }

    // Internal routing
    let router = Router::new()
        .get(handlers::index::index)
        // Child Router for authenticated endpoints
        .push(
            Router::new()
                // Hoops run a middleware before processing the actual request.
                .hoop(require_auth)
                .post(handlers::upload::upload)
                // File Management API
                .push(
                    Router::with_path("/api/v1/file/<filename>")
                        .delete(handlers::delete_file::delete_file), // .get(handlers::delete_file::delete_file), // TODO: serve RAW file, for programs. (when we add text/linting whatever)
                )
                // User Management Dashboard
                .push(Router::with_path("/profile").get(handlers::user_profile::user_profile))
                // User Management API
                .push(
                    Router::with_path("/api/v1/profile/generate_key")
                        .post(handlers::user_profile::generate_key),
                )
                // Admin Dashboard
                .push(Router::with_path("/admin").get(handlers::admin::admin))
                // Admin API
                .push(
                    Router::with_path("/api/v1/admin/<apikey>")
                        .delete(handlers::admin::admin_revoke_key)
                        .patch(handlers::admin::admin_edit_comment),
                ),
        )
        // Static Pages
        .push(Router::with_path("/services").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/about").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/czb").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/qr").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/faq").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/dmca").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/welcome").get(handlers::serve_static::serve_static))
        .push(Router::with_path("/metrics").get(handlers::serve_metrics::serve_metrics))
        // Might want to actually make a .ico instead of serving a webp image.
        .push(Router::with_path("/favicon.ico").get(StaticFile::new("static/favicon32.webp")))
        .push(Router::with_path("/robots.txt").get(StaticFile::new("static/robots.txt")))
        // Static File Serving (show auto index, which should be an option instead....)
        .push(Router::with_path("static/<**path>").get(StaticDir::new("static/").auto_list(true)))
        // Serving uploaded files, or 404.
        .push(Router::with_path("<file>").get(handlers::serve_file::serve_file));

    // Add OpenAPI Stuff here
    let doc = OpenApi::new("Ephemeral API", "0.1.0").merge_router(&router);
    let router = router
        .unshift(doc.into_router("/api-doc/openapi.json"))
        .unshift(SwaggerUi::new("/api-doc/openapi.json").into_router("/swagger-ui"));

    // Read environment variables for host and port
    let host = &CONFIG.server.host;
    let port = CONFIG.server.port;
    let server_url = format!("{}:{}", host, port);
    info!("Listening on http://{}", server_url);
    // TODO: Allow us to listen on a socket too?
    // let acceptor = UnixListener::new("/tmp/salvo.sock").bind().await;
    let acceptor = TcpListener::new(&server_url).bind().await;

    Server::new(acceptor).serve(router).await;

    // Close SQLite before closing
    info!("Attempting to safely shut down Ephemeral.");
    SQLITE
        .get()
        .expect("Problem while safely closing the database :(")
        .close()
        .await;
}
